@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: approve.c,v 1.108 1999/05/09 18:55:02 will Exp $";
+static char rcsid[] = "$Id: approve.c,v 1.135 2000/09/13 15:17:26 will Exp $";
 
 /* The code should compile with either ANSI C or K&R compilers. */
 
@@ -51,11 +51,17 @@ int verbosity;		/* HELP_BASIC, HELP_FULL, or HELP_FACTS */
     if (givehelp && verbosity == HELP_USAGE)
 	return "";
 
-    currfile = file_open(currfile, superfile, 0);
-    if (!currfile)
+    /* Open the super.tab file we are supposed to process, and then... */
+    if (!(currfile = file_open(NULL, superfile, 0, NULL, NULL)))
+	return NULL;
+    /* ...immediately chain-open the common super file (if it exists), so
+     * that it is always invoked by all commands (super.tab or per-user),
+     * and require it to be root-owned.
+     */
+    if (!(currfile = file_open(currfile, superfile_init, 1, 0, NULL)))
 	return NULL;
 
-    for (localinfo.progs.commandfound=NULL, currfile->line=1;
+    for (localinfo.progs.commandfound=NULL, currfile->line = 1;
 	(buf=fillbuffer(currfile->fp, &indentok, &currfile->nl)) ||
 	    ( (check_syntax || indentok) && currfile->prev != NULL);
 					     currfile->line += currfile->nl) {
@@ -259,8 +265,9 @@ CmdPattern or FullPathName missing.\n");
 
     /* Build lists of cmd::file entries */
     if ((i=build_cmd_file(usrcmd, givehelp || checksyntax,
-					word1, &cond_or_opt_wd)) <= 0)
+					word1, &cond_or_opt_wd)) <= 0) {
 	return i;
+    }
 
 
     if (!(givehelp || checksyntax)) {
@@ -271,8 +278,9 @@ CmdPattern or FullPathName missing.\n");
 
 	/* Adjust fullpath as necessary */
 	if (fixup_fullpath(indx, usrcmd, path,
-					FullPath, sizeof(FullPath)) == -1)
+					FullPath, sizeof(FullPath)) == -1) {
 	    return -1;
+	}
 
     } else {
 	/* givehelp or check syntax */
@@ -294,18 +302,22 @@ CmdPattern or FullPathName missing.\n");
     }
 
     /* Now process command-line conditions and options */
-    if ((condition_match=conditions_and_options(cond_or_opt_wd)) == -1)
+    if ((condition_match=conditions_and_options(cond_or_opt_wd)) == -1) {
 	return condition_match;
+    }
 
 
     if (condition_match && localinfo.die) {
-	if (checksyntax)
+	if (checksyntax || givehelp) {
 	    return 0;
-	else if (givehelp)
+	} else {
+	    extern char *error_prog;
+	    char *save_prog = error_prog;
+	    error_prog = NULL;
+	    Error(0, 0, "%s\n", strdup(do_variables(localinfo.die)));
+	    error_prog = save_prog;
 	    return -1;
-	else
-	    return Error(0, 0, "User's command: %s. %s\n",
-				    usrcmd, localinfo.die);
+	}
     }
     return condition_match;
 }
@@ -571,7 +583,10 @@ char *cond_or_opt_wd;	/* First word to process; rest we get with strqtokS */
 	    localinfo.userpats.next ||
 	    globalinfo.userbefore.next) {
 	/* We have user fields to match */
-	matches.user = 0;	/* default is to disallow user */
+
+	/* Default for root is to allow; default for all others is disallow */
+	matches.user = (userinfo.caller.pw_uid == 0) ? 1 : 0;
+
 	match_ugh_user(&globalinfo.userafter, 1);
 	if (!matches.user) {
 	    match_ugh_user(&localinfo.userpats, 0);
@@ -632,20 +647,25 @@ int verbosity;
 char *usrcmd;
 {
     static char *usage[] = {
-"    %s command [args...]	\texecutes command [args]",
-"    %s                         \tlist ok cmds",
-"    %s [-H | -f] [MasqueradeOpts]\tlist ok cmds: -H=verbose,-f=just-facts",
-"    %s -h			\tprint this usage info, then exits",
-"    %s -V			\tprint version number, then exits",
-"    %s [-d | -D] [MasqueradeOpts]\tcmd [args]\t...debug mode",
-"    %s -b			\tlists builtin variables, then exits",
-"    %s -c [superfile]	\tjust check syntax of the superfile",
+"",
+"  %s command [args...]   ......   executes command [args].",
+"",
+"  %s   ........................   basic listing of permitted cmds.",
+"  %s [-H | -f] [MasqueradeOpts]   list ok cmds: -H=verbose,-f=just-facts.",
+"  %s [-H | -f] [MasqueradeOpts] cmd  ...give help only for `cmd'.",
+"  %s -h   .....................   print this usage info, then exits.",
+"",
+"  %s -b   .....................   lists builtin variables, then exits.",
+"  %s -c [superfile]   .........   just check syntax of the superfile.",
+"  %s -V   .....................   print version number, then exits.",
+"  %s [-d | -D] [MasqueradeOpts] cmd [args]   ...debug mode.",
 "",
 "Notes on options:",
 "    -d   Basic debug: show what would be done, but don't execute anything.",
 "    -D   Same as -d, plus tell more about variables, plus other details.",
 "    -f   (just the facts, ma'm): lists your cmds as: Cmd FullPath [args]",
 "	      Useful for scripts that want a list of what you may execute.",
+"",
 "MasqueradeOptions:",
 "    ** No command is ever executed when any MasqueradeOption is used.**",
 "    Additionally, some debugging output is enabled when these are used:",
@@ -720,10 +740,11 @@ int verbosity;
 	     * otherwise just the command name.
 	     */
 	    if (strchr(localinfo.progs.cmd_file[0].Cmd, '*')) {
-		printf("%s -> %s:\n                -- %s\n",
+		printf("%s -> %s:\n",
 			localinfo.progs.cmd_file[0].Cmd,
-			localinfo.progs.cmd_file[0].File,
-			localinfo.info);
+			localinfo.progs.cmd_file[0].File);
+		if (localinfo.info)
+		    printf("\t\t-- %s\n", localinfo.info);
 	    } else {
 		printf("%-15s %s\n", localinfo.progs.cmd_file[0].Cmd,
 			localinfo.info ? localinfo.info : "");
@@ -822,7 +843,7 @@ int verbosity;
 	    localinfo.fdlist ||
 	    *localinfo.owner ||
 	    localinfo.env) {
-	(void) fputs("\t(Executes with:", stdout);
+	(void) fputs("\tExecutes with:", stdout);
 	if (*localinfo.user) (void) printf(" uid=%s", localinfo.user);
 	if (*localinfo.group) (void) printf(" gid=%s", localinfo.group);
 	if (*localinfo.u_g) (void) printf(" u+g=%s", localinfo.u_g);
@@ -836,7 +857,7 @@ int verbosity;
 	}
 	if (localinfo.fdlist) (void) printf(" fdlist=%s", localinfo.fdlist);
 	if (*localinfo.owner) (void) printf(" owner=%s", localinfo.owner);
-	(void) fputs(")\n", stdout);
+	(void) fputs("\n", stdout);
     }
 
     if (localinfo.usr_args[0] < 0)
@@ -878,14 +899,24 @@ int verbosity;
 	printf("\tCommand executes with umask set to 0%o.\n", rcl_umask());
 
 
-    if (localinfo.passinfo.required) {
-	if (localinfo.passinfo.timeout == 0)
-	    (void) printf("\tPassword is required on each use.\n");
+    if (localinfo.authinfo.required) {
+	if (localinfo.authinfo.timeout == 0)
+	    (void) printf("\t%s authentication is required on each use.\n",
+		auth_name());
 	else
-	    (void) printf("\tPassword is required; good for %d minutes%s.\n",
-		localinfo.passinfo.timeout, localinfo.passinfo.renewtime ? 
+	    (void) printf(
+		"\t%s authentication is required; good for %d minutes%s.\n",
+		auth_name(),
+		localinfo.authinfo.timeout, localinfo.authinfo.renewtime ? 
 		    ", extended with each use" : "");
     }
+    if (localinfo.checkvar) {
+	(void) printf("\tCaller must enter check-variables:");
+	for (i=0; localinfo.checkvar[i]; i++) {
+	    printf("  %s", localinfo.checkvar[i]);
+	}
+	(void) putchar('\n');
+    }
     (void) putchar('\n');
 }
 
