@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: checks.c,v 1.70 1999/06/15 08:05:54 will Exp $";
+static char rcsid[] = "$Id: checks.c,v 1.147 2000/12/15 17:12:54 will Exp $";
 /* The code should compile with either ANSI C or K&R compilers. */
 
 /*
@@ -20,6 +20,21 @@ static char rcsid[] = "$Id: checks.c,v 1.70 1999/06/15 08:05:54 will Exp $";
 #define netgrp_h_compare(p, h) 0
 #endif
 
+
+#ifdef _HPUX_SOURCE
+
+#ifdef HAVE_ISCOMSEC
+    char *bigcrypt();
+    char *getpasswd();
+#else
+    /* If we don't have iscomsec() on this HP system, fake it -- otherwise
+     * we'll have to #ifdef the use of iscomsec(), and it's cleaner code
+     * if we don't.
+     */
+    int iscomsec() { return 0; }
+#endif
+#endif
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Check that an environment variable only includes allowed characters */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -30,6 +45,7 @@ char *name;	/* variable name to check (e.g. "TERM") */
 char *value;	/* contents of variable (e.g. "vt100") */
 char *pat;	/* pattern that value must match */
 {
+    int len;
 
     if (!value)
 	return -1;
@@ -50,6 +66,112 @@ char *pat;	/* pattern that value must match */
 	return Error(0, 0,
 	    "checkenv(): $%.100s (=%.100s) doesn't match pattern %-.500s.\n",
 			name, value, pat);
+
+    /* Limit the value to a reasonable length (MAXENVLEN chars) */
+    len = strlen(value);
+    if (len > MAXENVLEN) {
+	return Error(0, 0, "Imported environment variables may not \
+exceed %d chars; you passed a string of length %d!\n", MAXENVLEN, len);
+    }
+
+    return 0;
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Check the checkvar=name,... list.
+ * If the list isn't empty, use /dev/tty for input.
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Returns 0 if list ok; -1 otherwise.  */
+int
+check_var_value()
+{
+    int l, ntry, match;
+    char buf[500];
+    char *varname, *truevalue;
+    int ivar;
+    FILE *prompt_fp;
+    char *prompt_dev="/dev/tty";
+
+    /* This is implemented as an inefficient operation, but it doesn't
+     * matter because we _assume_ the number of variables to be small
+     * and it's only invoked once.
+     */
+
+    if (!localinfo.checkvar || !localinfo.checkvar[0])
+	return 0;	/* no checkvar list */
+
+    if (!(prompt_fp = fopen(prompt_dev, "r+"))) {
+	return Error(0, 0,
+	    "Can't open %s to test checkvar=name,...\n", prompt_dev);
+    }
+    if (localinfo.checkvar[1]) {
+	/* There must be more than one variable */
+	fprintf(prompt_fp, "%c%s needs you enter some variables \
+before proceeding...\n", toupper(prog[0]), prog+1);
+    } else {
+	fprintf(prompt_fp, "%c%s needs you to enter the %s variable \
+before proceeding.\n", toupper(prog[0]), prog+1, localinfo.checkvar[0]);
+    }
+
+    for (ivar=0; localinfo.checkvar[ivar]; ivar++) {
+	varname = localinfo.checkvar[ivar];
+	truevalue = get_variable(varname);
+
+	if (!truevalue) {
+	    /* Ouch, checkvar=xxx is used, but xxx isn't defined! */
+	    fclose(prompt_fp);
+	    return Error(0, 0,
+		"%t\n\tError in super.tab file: \
+`checkvar=%s' is used, but variable %s isn't defined!\n", varname, varname);
+	}
+
+	for (ntry=0, match=0; ntry < MAXTRY && !match; ntry++) {
+	    if (ntry == 0) {
+		fprintf(prompt_fp, "Enter %s (<Return> for reminder): ",
+								    varname);
+	    } else {
+		fprintf(prompt_fp, "Enter %s (expecting `%s'): ",
+	    						varname, truevalue);
+	    }
+	    fflush(prompt_fp);
+	    if (!fgets(buf, sizeof(buf), prompt_fp)) {
+		if (feof(prompt_fp)) {
+		    fclose(prompt_fp);
+		    return Error(0, 0, "can't read %s -- stream was closed!\n",
+				prompt_dev);
+		} else if (ferror(prompt_fp)) {
+		    fclose(prompt_fp);
+		    return Error(0, 0, "error reading %s\n", prompt_dev);
+		} else {
+		    fclose(prompt_fp);
+		    return Error(0, 0, "??? feof() and ferror() return 0, \
+but I can't read %s\n", prompt_dev);
+		}
+	    }
+
+	    if (strlen(buf) == sizeof(buf)-1) {
+		fclose(prompt_fp);
+		return Error(0, 0, "Ridiculously long value <%.300s...> \
+returned for variable %s\n", buf, varname);
+	    }
+	    l = strlen(buf);
+	    if (buf[l-1] == '\n')
+		buf[l-1] = '\0';
+	    match = (strcmp(buf, truevalue) == 0);
+	    if (!match) {
+		fprintf(prompt_fp, "Variable %s incorrect\n", varname);
+	    }
+	}
+	if (!match) {
+	    fclose(prompt_fp);
+	    return Error(0, 0, "Never got variable %s entered correctly.\n",
+	    							varname);
+	}
+    }
+    fclose(prompt_fp);
+
+    /* all variables are ok. */
     return 0;
 }
 
@@ -167,7 +289,7 @@ requires owner to be %d (%s).\n",
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Gets an entry from the password file.  Optionally accepts special bracketed
- * name (e.g. <owner> or <caller>).  Accepts text uid's instead of names.
+ * name (e.g. <owner> or <caller>).  Accepts names or text uid's.
  * Returns ptr to password entry.  The password entry may be that returned
  * by getpwnam(), or it may be one already stored in a super-owned struct;
  * you can't make any assumptions about it -- therefore don't modify it.
@@ -222,6 +344,39 @@ char *username;		/* name to translate */
     return pw;
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Gets a group entry.
+ * Accepts text gid's or names.
+ * Returns ptr to group entry.
+ *
+ * The returned pointer points to an area that may be overwritten by later
+ * calls to the getgrxxx() routines; therefore if the caller wants to save
+ * the data, the data must be copied.
+ *
+ * On error, print message and return NULL.
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+struct group *
+getgrentry(name)
+char *name;		/* name to translate */
+{
+    struct group *gr;
+    char c;
+    int i, numeric;
+
+    gr = getgrnam(name);
+    if (!gr) {
+	numeric = (sscanf(name, "%d%c", &i, &c) == 1);
+	if (numeric)
+	    gr = getgrgid(i);
+    }
+    if (!gr) {
+	Error(0, 0, "%t\n\tNo such group or gid as `%s'.\n", name);
+	return NULL;
+    }
+    return gr;
+}
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Set supplementary groups according to the specified args                */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -353,13 +508,6 @@ set_u_g()
 	if (setuid(userinfo.new_uid) == -1)
 	    return Error(1, 0, "setuid(uid=%d) failed: ", userinfo.new_uid);
 
-    if (localinfo.passinfo.required) {
-	/* Get caller's encrypted password */
-	i = get_encrypted_pw();
-	if (i != 0)
-	    return i;
-    }
-
     return 0;
 }
 
@@ -387,32 +535,47 @@ get_encrypted_pw()
 
 #else
 #ifdef _HPUX_SOURCE
+
 int
 get_encrypted_pw()
 {
-    struct passwd *caller_pw;
+    if(iscomsec()) {
 
-    /* See if we can do shadow password lookup for HPUX 9.x.
-     * The rule is that if /.secure/etc/passwd exists, we have to use it;
-     * otherwise, fall through to regular password file lookup.
-     */
-    static struct stat st;
-    if (stat("/.secure/etc/passwd", &st) == 0) {
-	/* Shadow password file exists; use it */
-	struct s_passwd *caller_pw;
-	if (!(caller_pw = getspwnam(userinfo.caller.pw_name)))
+	struct pr_passwd *caller_pw;
+
+	if (!(caller_pw = getprpwnam(userinfo.caller.pw_name)))
 	    return Error(1, 0,
-		"Failed to obtain shadow password entry for user %s: ",
+		"Failed to obtain Trusted Computing Base entry for user %s: ",
 		userinfo.caller);
-	strcpy(userinfo.encr, caller_pw->pw_passwd);
-	strncpy(userinfo.salt, caller_pw->pw_passwd, 2);
+	strcpy(userinfo.encr, caller_pw->ufld.fd_encrypt);
+	strncpy(userinfo.salt, caller_pw->ufld.fd_encrypt, 2);
 	userinfo.salt[2] = '\0';
-    } else {
 
-	/* Fall through to regular password file lookup. */
-	strcpy(userinfo.encr, userinfo.caller.pw_passwd);
-	strncpy(userinfo.salt, userinfo.caller.pw_passwd, 2);
-	userinfo.salt[2] = '\0';
+    } else {
+	struct passwd *caller_pw;
+
+	/* See if we can do shadow password lookup for HPUX 9.x.
+	 * The rule is that if /.secure/etc/passwd exists, we have to use it;
+	 * otherwise, fall through to regular password file lookup.
+	 */
+	static struct stat st;
+	if (stat("/.secure/etc/passwd", &st) == 0) {
+	    /* Shadow password file exists; use it */
+	    struct s_passwd *caller_pw;
+	    if (!(caller_pw = getspwnam(userinfo.caller.pw_name)))
+		return Error(1, 0,
+		    "Failed to obtain shadow password entry for user %s: ",
+		    userinfo.caller);
+	    strcpy(userinfo.encr, caller_pw->pw_passwd);
+	    strncpy(userinfo.salt, caller_pw->pw_passwd, 2);
+	    userinfo.salt[2] = '\0';
+	} else {
+
+	    /* Fall through to regular password file lookup. */
+	    strcpy(userinfo.encr, userinfo.caller.pw_passwd);
+	    strncpy(userinfo.salt, userinfo.caller.pw_passwd, 2);
+	    userinfo.salt[2] = '\0';
+	}
     }
     return 0;
 }
@@ -451,6 +614,40 @@ get_encrypted_pw()
     return 0;
 }
 
+#else
+#ifdef Digital_UNIX
+int
+get_encrypted_pw()
+{
+    struct passwd *caller_pw;
+    struct pr_passwd *caller_prpass;
+
+    if (!(caller_pw = getpwnam(userinfo.caller.pw_name)))
+	return Error(0, 0, "No password entry for user %s.\n",
+		userinfo.caller.pw_name);
+
+    /* Digital Unix has "x" in password field to indicate shadow password
+     * file has to be consulted.
+     */
+    if (strcmp(caller_pw->pw_passwd, "x") == 0) {
+	/* Protected (shadow) password in use... */
+	if (!(caller_prpass = getprpwnam(userinfo.caller.pw_name)))
+	    return Error(1, 0,
+	    "Failed to obtain shadow password entry for user %s: ",
+	    userinfo.caller.pw_name);
+	strcpy(userinfo.encr, caller_prpass->ufld.fd_encrypt);
+	strncpy(userinfo.salt, caller_prpass->ufld.fd_encrypt, 2);
+	userinfo.salt[2] = '\0';
+
+    } else {
+	/* Fall through to regular password file lookup. */
+	strcpy(userinfo.encr, caller_pw->pw_passwd);
+	strncpy(userinfo.salt, caller_pw->pw_passwd, 2);
+	userinfo.salt[2] = '\0';
+    }
+    return 0;
+}
+
 #else
 #ifdef __linux__
 
@@ -493,10 +690,34 @@ get_encrypted_pw()
 #endif
 #endif
 #endif
+#endif
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-/* Checks if password is needed, and if so, prompts for same.
+/* Returns ptr to string with name of desired authentication method. */
+char *
+auth_name()
+{
+    if (!localinfo.authinfo.required) {
+	return "None";
+    } else if (localinfo.authinfo.method == SUPER_AUTH_PASSWORD) {
+	return "Password";
+    } else if (localinfo.authinfo.method == SUPER_AUTH_PAM) {
+	return "PAM";
+    } else {
+	Error(0, 1, "auth_name: don't recognize auth method %d!\n",
+						localinfo.authinfo.method);
+    }
+    /* NOTREACHED */
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Checks if authentication is needed, and does so if needed.
+ * Expects that the encrypted password is already in userinfo.{encr,salt},
+ * so that this can be run after switching to unpriviledged uid.
  * Returns 0 on success, -1 on error.
 
  * The timestamp directory faces the same problem as the logfile: if the
@@ -510,18 +731,19 @@ get_encrypted_pw()
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
-check_pass(cmd)
+check_auth(cmd)
 char *cmd;
 {
     char file[MAXPATHLEN];
     struct stat st;
-    int l, istat, file_exists, got_pass;
+    int l, istat, file_exists, got_auth;
     int timed_out = 0;
     int status;
     pid_t pid, child;
+    char *authuser;
 
-    if (!localinfo.passinfo.required)
-	return 0;			/* don't need password */
+    if (!localinfo.authinfo.required)
+	return 0;			/* don't need authentication */
 
     /* Create or update the timestamp file even if the lifetime is 0
      * (always ask for password).  We do this because the user may
@@ -551,121 +773,190 @@ expected child pid = %d; waiting for correct pid...\n", pid, child);
 	else if (status == 0)
 	    return 0;
 	else
-	    return Error(0, 0, "Timestamp creation failed\n");
+	    return Error(0, 0, "Authentication or timestamp creation failed\n");
+	/* NOTREACHED */
+    }
 	
-    } else {
-	/* In child.  setuid, then make and/or test the directory */
-	if (*localinfo.passinfo.user != '\0') {
-	    stringcopy(localinfo.user, localinfo.passinfo.user,
-						sizeof(localinfo.user));
-	    *localinfo.group = '\0';
-	    *localinfo.u_g = '\0';
-	    if (set_u_g() == -1) {
-		return Error(1, 0,
-		"failed to setuid %s before setting timestamp file: ",
-		localinfo.user);
-	    }
-	}
-	/* Make the timestamp directory name */
-	if (!makedirname(globalinfo.passinfo.perhost ? TIMESTAMP_DIR : "",
-						userinfo.hostname, file))
-	    return -1;
+    /*
+     * If here, must be in child.
+     */
+    if (child != 0) Error(0, 1,
+    			"Internal error: child=%d; should be 0!\n", child);
 
-	/* Make the timestamp directory */
-	if (makedir(file) == -1)
-	    return -1;
-
-	/* Make the file in the timestamp directory */
-	l = strlen(file) + 1 + strlen(userinfo.caller.pw_name);
-	if (l >= MAXPATHLEN)
+    /*
+     * setuid, then make and/or test the directory
+     */
+    if (*localinfo.authinfo.ts_user != '\0') {
+	stringcopy(localinfo.user, localinfo.authinfo.ts_user,
+					    sizeof(localinfo.user));
+	*localinfo.group = '\0';
+	*localinfo.u_g = '\0';
+	if (set_u_g() == -1) {
 	    return Error(1, 0,
-	    "Can't create timestamp file: it would exceed MAXPATHLEN = %d\n",
-	    MAXPATHLEN);
-	strcat(file, "/");
-	strcat(file, userinfo.caller.pw_name);
-
-	istat = stat(file, &st);
-	if (istat != 0 && errno != ENOENT)
-	    return Error(1, 0, "Failed to stat timestamp file `%s': ", file);
-
-	file_exists = (istat == 0);
-	if (file_exists)
-	    timed_out = (localinfo.passinfo.timeout < 1) ||
-			((time(NULL)-st.st_mtime) >
-			localinfo.passinfo.timeout*60);
-
-	got_pass=0;
-	if (!file_exists || timed_out) {
-	    got_pass = (get_password(cmd, userinfo.caller.pw_name,
-					userinfo.salt, userinfo.encr) == 1);
-	    if (!got_pass)
-		return -1;
+	    "failed to setuid %s before setting timestamp file: ",
+	    localinfo.user);
 	}
+    }
+    /* Make the timestamp directory name */
+    if (!makedirname(TIMESTAMP_DIR,
+		    globalinfo.authinfo.perhost ? userinfo.hostname : "",
+		    file))
+	return -1;
 
-	/* NOTE: A race condition is possible between two super's, with the
-	 * worst-case effect of an error message and failure to run the
-	 * requested command.
-	 */
+    /* Make the timestamp directory */
+    if (makedir(file) == -1)
+	return -1;
 
-	/* If file exists, and we haven't (a) gotten the password again, or
-	 * (b) supposed to automatically refresh the timestamp, do nothing to
-	 * the file except ensure that we own it.
+    /* Make the file in the timestamp directory */
+    l = strlen(file) + 1 + strlen(userinfo.caller.pw_name);
+    if (l >= MAXPATHLEN)
+	return Error(1, 0,
+	"Can't create timestamp file: it would exceed MAXPATHLEN = %d\n",
+	MAXPATHLEN);
+    strcat(file, "/");
+    strcat(file, userinfo.caller.pw_name);
 
-	 * Otherwise create the file (unlink it first if it exists).
-	 */
-	if (file_exists && !(got_pass || localinfo.passinfo.renewtime)) {
-	    if (st.st_uid != geteuid())
-		return Error(0, 0,
-	    "Timestamp file `%s' is owned by uid=%d, but expected owner=%d.\n\
+    istat = stat(file, &st);
+    if (istat != 0 && errno != ENOENT)
+	return Error(1, 0, "Failed to stat timestamp file `%s': ", file);
+
+    file_exists = (istat == 0);
+    if (file_exists)
+	timed_out = (localinfo.authinfo.timeout < 1) ||
+		    ((time(NULL)-st.st_mtime) >
+		    localinfo.authinfo.timeout*60);
+
+    got_auth=0;
+    if (localinfo.authinfo.user[0]) {
+	authuser = localinfo.authinfo.user;	/* authuser=xxx local opt */
+    } else if (globalinfo.authinfo.user[0]) {
+	authuser = globalinfo.authinfo.user;	/* authuser=xxx global opt */
+    } else {
+	authuser = userinfo.caller.pw_name;	/* default: caller's pw */
+    }
+    if (!file_exists || timed_out) {
+	switch (localinfo.authinfo.method) {
+	case SUPER_AUTH_PASSWORD:
+	    got_auth = (get_password(cmd, userinfo.caller.pw_name, authuser,
+				    userinfo.salt, userinfo.encr) == 1);
+	    break;
+	case SUPER_AUTH_PAM:
+	    got_auth = (get_pam(cmd, userinfo.caller.pw_name, authuser) == 1);
+	    break;
+	default:
+	    Error(0, 1, "Internal error in check_auth: \
+unknown auth method %d!\n", localinfo.authinfo.method);
+	}
+	if (!got_auth)
+	    return -1;
+    }
+
+    /* NOTE: A race condition is possible between two super's, with the
+     * worst-case effect of an error message and failure to run the
+     * requested command.
+     */
+
+    /* If file exists, and we haven't (a) gotten the password again, or
+     * (b) supposed to automatically refresh the timestamp, do nothing to
+     * the file except ensure that we own it.
+
+     * Otherwise create the file (unlink it first if it exists).
+     */
+    if (file_exists && !(got_auth || localinfo.authinfo.renewtime)) {
+	if (st.st_uid != geteuid())
+	    return Error(0, 0,
+	"Timestamp file `%s' is owned by uid=%d, but expected owner=%d.\n\
 \tN.B. If you recently changed the value of timestampuid=xxx, all existing\n\
 \tfiles in the timestamp directory _may_ have the wrong owner; delete them.\n\
 \t(No security hole appears when you delete a timestamp file.)\n",
-		file, st.st_uid, geteuid());
+	    file, st.st_uid, geteuid());
 
-	} else {
-	    if (file_exists) {
-		if (unlink(file) != 0)
-		    return Error(1, 0,
-			"Failed to unlink() timestamp file `%s': ", file);
-	    }
-	    if (open(file, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0200) == -1)
+    } else {
+	if (file_exists) {
+	    if (unlink(file) != 0)
 		return Error(1, 0,
-		    "Failed to open() timestamp file `%s': ", file);
+		    "Failed to unlink() timestamp file `%s': ", file);
 	}
-	exit(0);
+	if (open(file, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0200) == -1)
+	    return Error(1, 0,
+		"Failed to open() timestamp file `%s': ", file);
     }
+    exit(0);
     /* UNREACHABLE */
     Error(0, 1, "Unreachable code!\n");
     return 0;
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Encrypt a password.  Returns value from crypt() or similar. */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+char *
+docrypt(prompt, salt)
+char *prompt;
+char *salt;
+{
+    char *crypt();
+    char *getpass();
+
+#ifdef _HPUX_SOURCE
+
+    if (iscomsec()) {
+	#if (HPUX_MAJOR == 10)
+	    return bigcrypt(getpasswd(prompt), salt);
+	#else
+	    return crypt(getpass(prompt), salt);
+	#endif
+    } else {
+	return crypt(getpass(prompt), salt);
+    }
+
+#else
+
+    return crypt(getpass(prompt), salt);
+
+#endif
+}
+
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Gets a user's encrypted password. Returns -1 on failure, +1 on success */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
-get_password(cmd, user, salt, encr)
-char *cmd, *user, *salt, *encr;
+get_password(cmd, caller, user, salt, encr)
+char *cmd;
+char *caller;	/* the person who invoked super */
+char *user;	/* the person whose authentication is required */
+char *salt;
+char *encr;
 {
     /* No such file or password timed out -- get password */
     int ntry, match;
     char msg[500];
-    char *crypt(), *getpass();
     char *encrypted = NULL;
+
     if (strcmp(encr, "") == 0) {
 	return Error(0, 0,
-	    "Command requires a password, but user `%s' has no password\n",
-    user);
+	"Command requires a password, but user `%s' has no password\n", user);
     }
     for (ntry=0, match=0; ntry < MAXTRY && !match; ntry++) {
 	if (ntry == 0) {
-	    (void) sprintf(msg,
-    "Your password is required for super command `%s'...\nPassword: ",
+	    if (localinfo.authinfo.prompt && localinfo.authinfo.prompt[0]) {
+		stringcopy(msg,
+			do_variables(localinfo.authinfo.prompt), sizeof(msg));
+	    } else if (strcmp(caller, user) == 0) {
+		(void) sprintf(msg,
+	"Your password is required for super command `%.400s'...\nPassword: ",
 		cmd);
+	    } else {
+		(void) sprintf(msg,
+	"%c%s's password is required for super command `%.400s'...\nPassword: ",
+		toupper(*user), user+1, cmd);
+	    }
 	} else {
 	    strcpy(msg, "Password incorrect\nPassword: ");
 	}
+        encrypted = docrypt(msg, salt);
 
-	encrypted = crypt(getpass(msg), salt);
 	if (encr && encrypted) {
 	    match = (strcmp(encr, encrypted) == 0);
 	} else {
@@ -989,9 +1280,6 @@ char *token;		/* user/group/host pattern */
 		sizeof(chkbuf)-4);
     }
 
-    if (userinfo.caller.pw_uid == 0)
-	return 0;			/* root is always legit */
-
     /* Split into user:group@host; check host part first (if it exists) */
     if ((hostpat = strchr(token, '@'))) {
 	if (hostpat[1] == 0)
