@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: colon.c,v 1.19 1998/03/02 06:53:59 will Exp $";
+static char rcsid[] = "$Id: colon.c,v 1.28 2000/02/07 17:22:06 will Exp $";
 
 #include "super.h"
 #include "version.h"
@@ -31,6 +31,16 @@ int *continue_after_if;	/* return !0 iff a :if command evaluates true,
 	/* Process variable definition */
 	return colon_define(command);
 
+    } else if (strcmp(":die", command) == 0) {
+
+	/* Force immediate exit */
+	return colon_die(command);
+
+    } else if (strcmp(":getenv", command) == 0) {
+
+	/* Define variable from envvar */
+	return colon_getenv(command);
+
     } else if (strcmp(":include", command) == 0) {
 
 	/* Process include-file directive; don't allow missing files */
@@ -252,6 +262,62 @@ variable name missing after ':define'.\n");
     return add_variable(varname, varbody);
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Process the :die command */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+int
+colon_die(command)
+char *command;
+{
+    char *text;
+    extern char *strqS_start;
+
+#if 1
+    /* This collects next token only */
+    text = strqtokS(NULL, SEP, NULL, NULL, 1);
+#else
+    /* This collects all tokens (but comments have been elided); however,
+     * quotes remain in text.
+     */
+    text = strqS_start;
+#endif
+    Error(0, 1, "%s\n", text);
+
+    /* shouldn't ever get here */
+    return -1;
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Process the :getenv command */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+int
+colon_getenv(command)
+char *command;
+{
+    extern char *strqS_start;
+    char *varname, *varbody;
+    char *s, *defn;
+    int l;
+
+    for ( ; (varname = strqtokS(NULL, SEP, NULL, NULL, 1)) ; ) {
+	defn = getenv(varname);
+	if (!defn) {
+	    defn = "";
+	} else if (checkenv(varname, defn, "^[-/:+._a-zA-Z0-9]*$") == -1) {
+	    /* Record the problem but continue. */
+	    Error(0,0, "%t\n\timproper definition of envvar `%s': `%-.500s'\n",
+	    			varname, defn);
+	    continue;
+	}
+	/* Store value */
+	if (add_variable(varname, defn) != 0) {
+	    /* Better return -- we couldn't store a definition */
+	    return -1;
+	}
+    }
+    return 0;
+}
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Process the :if command */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -302,15 +368,60 @@ colon_include(command, allow_missing)
 char *command;
 int allow_missing;	/* don't complain if file is missing */
 {
-    char *filename, *wd, *s;
+    char *filename, *wd, *s, *opt;
     extern char *strqS_start;
+    struct passwd *uid_pw;
+    struct group *gr;
     FileList *fl;
+    char *owner = NULL;	/* xxx from owner=xxx */
+    char *group = NULL;	/* xxx from group=xxx */
+    uid_t uid;		/* uid of owner=xxx; -1 means no owner=xxx is set */
+    gid_t gid;		/* gid of group=xxx; -1 means no group=xxx is set */
 
     filename = strqtokS(NULL, SEP, NULL, NULL, 1);
     if (!filename)
 	return Error(0, 0, "%t\n\tformat error in super.tab file: \
 filename missing after ':include'.\n");
 
+    /* Check for owner= and group= options */
+    while ((opt = strqtokS(NULL, SEP, NULL, NULL, 1))) {
+	/* Check for and delete possible newline from opt */
+	s = opt + strlen(opt)-1;
+	if (s >= opt && *s == '\n')
+	    *s = '\0';
+	if (strncmp(opt, "owner=", 6) == 0) {
+	    owner = opt + 6;
+	    if (*owner == '\0') {
+		return Error(0, 0,
+			"%t\n\towner not specified in `owner=' option");
+	    }
+	} else if (strncmp(opt, "group=", 6) == 0) {
+	    group = opt + 6;
+	    if (*group == '\0') {
+		return Error(0, 0,
+			"%t\n\tgroup not specified in `group=' option");
+	    }
+	} else {
+	    return Error(0, 0,
+	    	"%t\n\tgarbage text <%s> after :include command", opt);
+	}
+    }
+
+    if (owner) {
+	/* :include line contained owner= entry */
+	if (!(uid_pw = getpwentry(0, owner))) {
+	    return -1;
+	}
+	uid = uid_pw->pw_uid;
+    }
+    if (group) {
+	/* :include line contained owner= entry */
+	if (!(gr = getgrentry(group))) {
+	    return -1;
+	}
+	gid = gr->gr_gid;
+    }
+
     wd = strqS_start;
     if (wd && *wd != '\n' && *wd != '\0')
 	return Error(0, 0, "%t\n\tformat error in super.tab file: \n\t\
@@ -321,7 +432,8 @@ extra text <%s> after :include filename\n", wd);
     if (s >= filename && *s == '\n')
 	*s = '\0';
 
-    fl = file_open(currfile, filename, allow_missing);
+    fl = file_open(currfile, filename, allow_missing,
+    				owner ? &uid : NULL, group ? &gid : NULL);
     if (!fl)
 	return -1;
 
