@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: error.c,v 1.64 1999/02/18 19:52:35 will Exp $";
+static char rcsid[] = "$Id: error.c,v 1.67 2000/08/20 05:36:55 will Exp $";
 
 /*
  *	Copyright (c) 1993 by California Institute of Technology.
@@ -50,6 +50,7 @@ static char rcsid[] = "$Id: error.c,v 1.64 1999/02/18 19:52:35 will Exp $";
  *	error_rlog_host: host to receive rsyslog() messages.  Def: ourselves.
  *	error_user: Controls username used when printing messages.
  *	error_tag: Controls use of error_{srcfile,line,nl}.  Default off.
+ *	error_fflush: fflush(stdout) before printing error message.  Default on.
  *	error_srcfile, error_line, error_nl: error_srcfile is a file in
  *			which an error was found, starting at line error_line
  *			and continuing over error_nl lines.
@@ -144,6 +145,7 @@ int error_line = -1;
 int error_nl = -1;
 char *error_srcfile = NULL;
 int error_tag = 0;
+int error_fflush = 1;
 
 #ifdef HAVE_SYSLOG_H
 
@@ -208,6 +210,9 @@ Error(
     char *tag;
     int pctt;
 
+    if (error_fflush)
+	fflush(stdout);
+
     error = errno;
 
     /* Figure out line tagging */
@@ -312,8 +317,8 @@ Error(
 #ifdef HAVE_SYSLOG_H
     if (error_syslog) {
 	char newfmt[MAXPRINT], buf[MAXPRINT];
-	if (!openlog_done && error_prog) {
-	    OpenLog(error_prog, 0, error_facility);
+	if (!openlog_done) {
+	    OpenLog(error_prog ? error_prog : "", 0, error_facility);
 	    openlog_done = 1;
 	}
 	sprintf(newfmt, "(%s) ", error_user ? error_user : user);
@@ -351,6 +356,9 @@ va_dcl
     int pctt;
     FILE *error_cmd = NULL;
 
+    if (error_fflush)
+	fflush(stdout);
+
     error = errno;
 
     /* Figure out line tagging */
