@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: options.c,v 1.86 1999/05/05 17:24:04 will Exp $";
+static char rcsid[] = "$Id: options.c,v 1.105 2000/09/15 21:50:53 will Exp $";
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Process a global xxx=yyy option */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -15,7 +15,12 @@ Option opts[] = {
 	{ "addgroups",		LIST|GLOBAL|LOCAL,	option_groups },
 	{ "arg",		LOCAL,		option_arg },
 	{ "argv0",		LOCAL,		option_argv0 },
+	{ "auth",		GLOBAL|LOCAL,	option_auth },
+	{ "authprompt",		GLOBAL|LOCAL,	option_authprompt },
+	{ "authtype",		GLOBAL|LOCAL,	option_authtype },
+	{ "authuser",		GLOBAL|LOCAL,	option_authuser },
 	{ "cd",			GLOBAL|LOCAL,	option_cd },
+	{ "checkvar",		LIST|LOCAL,	option_checkvar },
 	{ "die",		LOCAL,		option_die },
 	{ "env",		LIST|GLOBAL|LOCAL,	option_env },
 	{ "fd",			LIST|LOCAL,	option_fd },
@@ -37,6 +42,7 @@ Option opts[] = {
 	{ "password",		GLOBAL|LOCAL,	option_password },
 	{ "patterns",		GLOBAL,		option_patterns },
 	{ "print",		LOCAL,		option_print },
+	{ "pwprompt",		GLOBAL|LOCAL,	option_authprompt },
 	{ "relative_path",	GLOBAL,		option_relative_path },
 	{ "renewtime",		GLOBAL,		option_renewtime },
 	{ "rlog_host",		GLOBAL,		option_rlog_host },
@@ -317,14 +323,143 @@ option_local_clear_settings()
 
     localinfo.usr_args[0] = localinfo.usr_args[1] = -1;
     StrEltsUnused(&localinfo.argpats);
+    if (localinfo.checkvar)
+	blkfree(localinfo.checkvar);
+    localinfo.checkvar = NULL;
     /* should use struct assignment, but that's not portable to
      * all k&r compilers.
      */
-    localinfo.passinfo.required = globalinfo.passinfo.required;
-    localinfo.passinfo.timeout = globalinfo.passinfo.timeout;
-    localinfo.passinfo.renewtime = globalinfo.passinfo.renewtime;
-    localinfo.passinfo.perhost = globalinfo.passinfo.perhost;
-    strcpy(localinfo.passinfo.user, globalinfo.passinfo.user);
+    localinfo.authinfo.required = globalinfo.authinfo.required;
+    localinfo.authinfo.method = globalinfo.authinfo.method;
+    localinfo.authinfo.timeout = globalinfo.authinfo.timeout;
+    localinfo.authinfo.renewtime = globalinfo.authinfo.renewtime;
+    localinfo.authinfo.perhost = globalinfo.authinfo.perhost;
+    if (localinfo.authinfo.prompt)
+	free(localinfo.authinfo.prompt);
+    localinfo.authinfo.prompt = globalinfo.authinfo.prompt;
+    strcpy(localinfo.authinfo.user, globalinfo.authinfo.user);
+    strcpy(localinfo.authinfo.ts_user, globalinfo.authinfo.ts_user);
+
+    return 0;
+}
+
+/***************************************************************************/
+/*                                                                         */
+/***************************************************************************/
+static int
+option_auth(word, s, isglobal)
+char *word;
+char *s;
+int isglobal;
+{
+    int required;
+
+    if (debug)
+	fprintf(stderr, "\toption:auth=%s (%s)\n",
+				s, isglobal ? "global" : "local");
+
+    if (strcmp(s, "y") == 0) {
+	required = 1;
+    } else if (strcmp(s, "n") == 0) {
+	required = 0;
+    } else {
+	return Error(0, 0, "%t\n\tInvalid option value \
+in `%s' -- value must be `y' or `n'\n", word);
+    }
+
+    if (required && using_user_supertab) {
+	return Error(0, 0, "%t\n\tUser-supplied .supertab files may not do \
+authentication checking -- sorry.\n");
+    }
+
+    if (isglobal)
+	globalinfo.authinfo.required = required;
+    else
+	localinfo.authinfo.required = required;
+
+    return 0;
+}
+
+static int
+option_authprompt(word, s, isglobal)
+char *word;
+char *s;
+int isglobal;
+{
+    char *str;
+
+    if (debug)
+	fprintf(stderr, "\toption:authprompt=%s (%s)\n",
+				s, isglobal ? "global" : "local");
+
+    if (using_user_supertab) {
+	return Error(0, 0, "%t\n\tUser-supplied .supertab files may not do \
+password checking -- sorry.\n");
+    }
+
+    str = strdup(s);
+    if (!str)
+	return Error(0, 0,
+		    "%t\n\tfailed to malloc space for authprompt=<%s>\n", s);
+    if (isglobal)
+	globalinfo.authinfo.prompt = str;
+    else
+	localinfo.authinfo.prompt = str;
+
+    return 0;
+}
+
+static int
+option_authtype(word, s, isglobal)
+char *word;
+char *s;
+int isglobal;
+{
+    int method;
+
+    if (debug)
+	fprintf(stderr, "\toption:authtype=%s (%s)\n",
+				s, isglobal ? "global" : "local");
+
+    if (strcmp(s, "pam") == 0) {
+	method = SUPER_AUTH_PAM;
+    } else if (strcmp(s, "password") == 0) {
+	method = SUPER_AUTH_PASSWORD;
+    } else {
+	return Error(0, 0, "%t\n\tInvalid option value \
+in `%s' -- value must be `pam' or `password'\n", word);
+    }
+
+    if (isglobal)
+	globalinfo.authinfo.method = method;
+    else
+	localinfo.authinfo.method = method;
+
+    return 0;
+}
+
+static int
+option_authuser(word, s, isglobal)
+char *word;
+char *s;
+int isglobal;
+{
+    int method;
+
+    if (debug)
+	fprintf(stderr, "\toption:authuser=%s (%s)\n",
+				s, isglobal ? "global" : "local");
+
+    /* Don't do any logging if debugging or just checking syntax! */
+    if (check_syntax || it_came_from_cmdline) {
+	Error(0, 0, "%t\n\tIgnoring authuser option, due to debug mode\n");
+	return 0;
+    }
+
+    if (isglobal)
+	strcpy(globalinfo.authinfo.user, s);
+    else
+	strcpy(localinfo.authinfo.user, s);
 
     return 0;
 }
@@ -425,6 +560,26 @@ int isglobal;
     return 0;
 }
 
+static int
+option_checkvar(word, s, isglobal)
+char *word;
+char *s;
+int isglobal;
+{
+    if (debug)
+	fprintf(stderr, "\toption:checkvar=%s (%s)\n",
+				s, isglobal ? "global" : "local");
+
+    if (isglobal) {
+	return Error(0, 0,
+		    "internal error -- option_checkvar called as global!\n");
+    } else {
+	localinfo.checkvar = optionlist;
+    }
+    return 0;
+}
+
+
 static int
 option_die(word, s, isglobal)
 char *word;
@@ -693,9 +848,11 @@ int isglobal;
 	fprintf(stderr, "\toption:loguid=%s (%s)\n",
 				s, isglobal ? "global" : "local");
 
-    /* Don't do any logging if just checking syntax! */
-    if (check_syntax)
+    /* Don't do any logging if debugging or just checking syntax! */
+    if (check_syntax || it_came_from_cmdline) {
+	Error(0, 0, "%t\n\tIgnoring loguid option, due to debug mode\n");
 	return 0;
+    }
 
     strcpy(globalinfo.log.user, s);
     globalinfo.log.newuid = 1;
@@ -877,7 +1034,7 @@ char *word;
 char *s;
 int isglobal;
 {
-    int required;
+    int method, required;
 
     if (debug)
 	fprintf(stderr, "\toption:password=%s (%s)\n",
@@ -897,10 +1054,15 @@ in `%s' -- value must be `y' or `n'\n", word);
 password checking -- sorry.\n");
     }
 
-    if (isglobal)
-	globalinfo.passinfo.required = required;
-    else
-	localinfo.passinfo.required = required;
+    if (isglobal) {
+	globalinfo.authinfo.required = required;
+	if (required)
+	    globalinfo.authinfo.method = SUPER_AUTH_PASSWORD;
+    } else {
+	localinfo.authinfo.required = required;
+	if (required)
+	    localinfo.authinfo.method = SUPER_AUTH_PASSWORD;
+    }
 
     return 0;
 }
@@ -978,9 +1140,9 @@ int isglobal;
 				s, isglobal ? "global" : "local");
 
     if (strcmp(s, "y") == 0) {
-	globalinfo.passinfo.renewtime = 1;
+	globalinfo.authinfo.renewtime = 1;
     } else {
-	globalinfo.passinfo.renewtime = 0;
+	globalinfo.authinfo.renewtime = 0;
 	if (strcmp(s, "n") != 0)
 	    return Error(0, 0, "%t\n\tInvalid option value \
 in `%s' -- value must be `y' or `n'\n", word);
@@ -1154,9 +1316,9 @@ int isglobal;
 				s, isglobal ? "global" : "local");
 
     if (isglobal)
-	globalinfo.passinfo.timeout = atoi(s);
+	globalinfo.authinfo.timeout = atoi(s);
     else
-	localinfo.passinfo.timeout = atoi(s);
+	localinfo.authinfo.timeout = atoi(s);
 
     return 0;
 }
@@ -1172,9 +1334,9 @@ int isglobal;
 				s, isglobal ? "global" : "local");
 
     if (strcmp(s, "n") == 0) {
-	globalinfo.passinfo.perhost = 0;
+	globalinfo.authinfo.perhost = 0;
     } else if (strcmp(s, "y") == 0) {
-	globalinfo.passinfo.perhost = 1;
+	globalinfo.authinfo.perhost = 1;
     } else {
 	return Error(0, 0,
 		"%t\n\tInvalid option value in `%s' -- \
@@ -1193,7 +1355,7 @@ int isglobal;
 	fprintf(stderr, "\toption:timestampuid=%s (%s)\n",
 				s, isglobal ? "global" : "local");
 
-    strcpy(globalinfo.passinfo.user, s);
+    strcpy(globalinfo.authinfo.ts_user, s);
     return 0;
 }
 
