@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: super.c,v 1.486 1999/05/05 17:30:39 will Exp $";
+static char rcsid[] = "$Id: super.c,v 1.566 2000/09/15 21:49:42 will Exp $";
 /* The code should compile with either ANSI C or K&R compilers. */
 
 /*
@@ -25,14 +25,14 @@ static char rcsid[] = "$Id: super.c,v 1.486 1999/05/05 17:30:39 will Exp $";
  *     $0 [-H | -f] [-F superfile] [ -T hh[:mm][/day] ] [-U uid] [-G gid]
  *		...lists allowed commands.
  *
- *     $0 command [args...]
+ *     $0 [-o file] command [args...]
  *		...execute command.
  *
  *     $0 -b
  *		...lists builtin variables (useful when creating super.tab).
  *
- *     $0 [-d | -D] [-F superfile] [-T time] [-U uid] [-G gid] [-M mach] [--] \
- *						[commandname [args...]]
+ *     $0 [-d | -D] [-F superfile] [-T time] [-U uid] [-G gid] [-M mach] \
+ *			[-o file] [--] [commandname [args...]]
  *		...debug mode.
  *
  *     $0 -c [superfile]
@@ -70,12 +70,16 @@ static char rcsid[] = "$Id: super.c,v 1.486 1999/05/05 17:30:39 will Exp $";
 
  */
 
-char superfile_writable[MAXPATHLEN+100]
-			= SUPERFILE;	/* The super.tab file -- put it
+/* Non-STD-C implementations do not have uniform ways of pasting strings
+ * together.  So we will compose the actual superfile and super.init
+ * names at runtime.
+ */
+char superfile_writable[MAXPATHLEN+100];/* The super.tab file -- put it
 					 * into a char array so we can
 					 * overwrite it later, if nec.
 					 */
-char *superfile = superfile_writable;	/* The super.tab file */
+char *superfile = superfile_writable;
+char superfile_init[MAXPATHLEN+100];	/* The super.init file */
 
 /* Global info */
 GlobalInfo globalinfo = {
@@ -91,11 +95,14 @@ GlobalInfo globalinfo = {
 	MAXLENARGS,	/* maximum length of all arguments, combined */
 
 	 {		/* Global password requirements */
-		0,	/* required (bool) */
+		0,	/* bool: is authentication required? */
+		SUPER_AUTH_PASSWORD, /* authentication method, if required */
 		5,	/* timeout (min) */
 		0,	/* renewtime (bool) */
 		1,	/* perhost (bool) */
-		""},	/* user (user who owns file) */
+		"",	/* authuser */
+		"",	/* user (user who owns timestampuid file) */
+		NULL},	/* prompt */
 
 
 	{0,NULL,NULL},	/* userbefore: list of u/g/h pats before per-cmd pats */
@@ -165,6 +172,7 @@ int debug=0;				/* Set by the debug options flag */
 int it_came_from_cmdline=0;		/* Set by -F/-T/-U/-G/-M flags */
 int check_syntax=0;			/* Set by the -c options flag */
 int using_user_supertab = 0;		/* !0 means using a user's .supertab */
+SimpleList Var_Value = {NULL, NULL};
 
 /* Routines used to compile/match user/group/host patterns */
 char *(*pat_compile) P__((char *)) = s_re_comp;
@@ -185,7 +193,6 @@ unsigned char my_cc[256];		/* our comment characters */
     struct rlimit new_core_limit;
 #endif
 
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
@@ -194,10 +201,11 @@ int argc;
 char **argv;
 {
     int status, n_builtin;
-    char *s, *cmd, *path;
+    char *s, *cmd, *path, *user;
     char **arglist, **envp;
     extern char *error_prog;
     int iarg, givehelp, giveversion, verbosity, printvars;
+    char *o_file;	/* The argument to -o xxx */
 
 #ifdef RLIMIT_CORE
     getrlimit(RLIMIT_CORE, &old_core_limit);
@@ -205,12 +213,16 @@ char **argv;
     new_core_limit.rlim_max = old_core_limit.rlim_max;
     setrlimit(RLIMIT_CORE, &new_core_limit);
 #endif
+    sprintf(superfile_writable, "%s/%s", SUPERDIR, "super.tab");
+    sprintf(superfile_init, "%s/%s", SUPERDIR, "super.init");
 
     s = strrchr(argv[0], '/');
     prog = (s && *(s+1)) ? s+1 : argv[0];
     error_prog = ONETRUENAME;	/* same as prog, but used by Error() */
 
     debug = check_syntax = giveversion = givehelp = printvars = 0;
+    verbosity = HELP_BASIC;	/* only matters if *givehelp != 0 */
+    o_file = NULL;
 
     error_srcfile = superfile;	/* error messages w/ line nums refer to this */
     error_line = -1;
@@ -224,26 +236,61 @@ char **argv;
 
     add_builtin_variables();	/* Define variables that describe this system */
 
+    if (get_encrypted_pw()!=0) { /* Get caller's encrypted password while we */
+				/* definitely have privs.  This allows      */
+				/* .supertab files to do password checking. */
+	*userinfo.encr = '\0';
+	*userinfo.salt = '\0';
+	if (geteuid() != 0) {
+	    Error(0, 0, "\tWon't be able to do password-checking.  \
+Note: this copy of super is not \
+running setuid-root, so it can't check passwords on modern Unix systems.\n\n");
+	} else {
+	    Error(0, 0, "\tWon't be able to do authentication (password-checking, or whatnot).\n\n");
+	}
+    }
+
     /* Decide if we were invoked as "super cmd args...", or 
      * as a link: cmd args...
      */
     if (strcmp(prog, ONETRUENAME) == 0) {
-	/* Invoked as super [options] cmd args... */
+	/* Invoked as super [options] cmd [arg ...]. */
+
 	iarg = do_options(argc, argv, &givehelp, &giveversion,
-				&verbosity, &printvars);
+			&verbosity, &printvars, &o_file);
 	if (iarg < 0 || (argv[0] == NULL && !(givehelp || giveversion))) {
 	    /* User screwed up; give minimal help */
 	    fprintf(stderr, "Type   %s -h   for usage information.\n", prog);
 	    exit(0);
 	}
-	cmd = argv[iarg];
-	argv += iarg;
-	/* Check if there was no cmd given after options. */
-	if (!cmd || !*cmd) {
-	    if (!giveversion)		/* If user didn't give -V, and didn't */
-		givehelp = 1;		/* ...give cmd, default is givehelp */
-	    cmd = NULL;
+	if (o_file) {
+	    /* Cmd is last component of the o_file path, e.g.
+	     * -o a/b/c means the cmd is "c".
+	     */
+	    if ((cmd = strrchr(o_file, '/'))) {
+		/* found last slash in path */
+		cmd++;
+		if (*cmd == '\0')
+		    cmd = NULL;	/* shouldn't ever happen */
+	    } else {
+		/* there is no slash */
+		cmd = o_file;
+	    }
+	    argv += iarg - 1;	/* Skip over the super options */
+
+	} else {
+	    /* Cmd is next argument on cmd line, if any */
+	    cmd = argv[iarg];
+	    argv += iarg;	/* Skip over the super options */
+
+	    if (!cmd || !*cmd) {
+		/* No cmd was given after options. */
+		if (!giveversion)	/* If user didn't give -V, and didn't */
+		    givehelp = 1;	/* ...give cmd, default is givehelp */
+		cmd = NULL;
+	    }
 	}
+
 	/* Test if we simply print variables, then bail out. */
 	if (printvars) {
 	    printf("Builtin variables:\n");
@@ -263,19 +310,42 @@ char **argv;
 
     init_strqtokS();
 
+    /*
+     * Check if we need to switch to processing a user's super file
+     */
+    using_user_supertab = 0;
+    if (o_file) {
+	/* Have `-o file' */
+
+	using_user_supertab = 1;
+	user_supertab(o_file, 1, cmd);
+
+    } else if (cmd && (s=strchr(cmd, ':'))) {
+	/* Have user:command */
+
+	using_user_supertab = 1;
+	user = cmd;
+	if (s == user)
+	    Error(0, 1, "Commands may not begin with `:'\n");
+	*s = '\0';	/* null-terminate the `user' part */
+	cmd = s+1;
+	if (!*cmd) {
+	    /* No cmd given after `user:' */
+	    if (!giveversion)		/* If user didn't give -V, and didn't */
+		givehelp = 1;		/* ...give cmd, default is givehelp */
+	    cmd = NULL;
+	}
+	user_supertab(user, 0, cmd);
+    }
+    /* Set a variable that a super.tab (or super.init) can use to determine
+     * if we're processing a user's .supertab or not.
+     */
+    if (add_variable("IS_USERTAB", using_user_supertab ? "yes" : "no") == -1)
+	exit(1);
+
     if (check_syntax)
 	Error(0,0, "Checking syntax of superfile `%s'\n", superfile);
 
-    /* Check if we need to switch to processing a user's super file */
-    using_user_supertab = 0;
-    if (cmd && strchr(cmd, ':')) {
-	char *tcmd = strchr(cmd, ':') + 1;
-	if (user_supertab(cmd) == -1)
-	    return 1;
-	cmd = tcmd;
-	using_user_supertab = 1;
-    }
-
     /* Print version if explicitly requested or if giving help */
     if ((givehelp && verbosity != HELP_FACTS) || giveversion)
 	(void) printf("%s version %s patchlevel %s\n",
@@ -286,10 +356,11 @@ char **argv;
 	exit(0);
 
     /* Check for permission to execute, and change uid/gid as necessary */
-    if ((path = approve(cmd, givehelp, verbosity)) == NULL)
+    if ((path = approve(cmd, givehelp, verbosity)) == NULL) {
 	return 1;
-    else if (*path == '\0')
+    } else if (*path == '\0') {
 	return 0;
+    }
 
     /* Get the arglist for the command, and null-terminate cmd if not
      * already done so.  Do this before things like get_owner, because
@@ -331,6 +402,14 @@ char **argv;
 	}
     }
 
+    /* Check authorization requirements.
+     * Just as for negative nice increment, this require root privileges
+     * to read a shadow password file or switch to another uid for writing
+     * the timestamp file.
+     */
+    if (check_auth(cmd) != 0)
+	return 1;
+
     /* Set uid/gid if necessary */
     if (using_user_supertab) {
 	/* Already changed to user's uid/gid earlier */
@@ -372,10 +451,11 @@ due to previously-reported problem)\n");
 	return 1;
     }
 
-    /* Check password requirements */
-    if (check_pass(cmd) != 0)
+    /* Do checks required by a checkvar=a,b,c... list */
+    if (check_var_value() != 0)
 	return 1;
 
+
     /* Log an informational message at LOG_INFO priority, not at
      * the usual error priority.
      */
@@ -458,13 +538,15 @@ due to previously-reported problem)\n");
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int
-do_options(argc, argv, givehelp, giveversion, verbosity, printvars)
+do_options(argc, argv, givehelp, giveversion, verbosity,
+					printvars, o_file)
 int argc;
 char **argv;
 int *givehelp, *giveversion, *verbosity, *printvars;
+char **o_file;
 {
     int iarg;
-    char *s, *t;
+    char *mhost, *s, *t;
     short minutes;
 #ifndef HAVE_LOCALTIME
     char daynum;
@@ -473,8 +555,12 @@ int *givehelp, *giveversion, *verbosity, *printvars;
     long l;
     struct passwd *pw;
 
-    /* In case arguments don't have defaults already assigned */
-    debug = check_syntax = *giveversion = *givehelp = *printvars = 0;
+    /* NOTE WELL: THE CALLER IS EXPECTED TO HAVE INITIALIZED THE ARGUMENTS
+     * TO THE DESIRED DEFAULT VALUES.  It's not repeated here because
+     * this routine isn't called for super invoked via symlink, and
+     * it's misleading to give the impression that the options are always
+     * going to be init'd here.
+     */
     *verbosity = HELP_BASIC;	/* only matters if *givehelp != 0 */
 
     for (iarg = 1; iarg < argc && argv[iarg][0] == '-'; iarg++) {
@@ -524,26 +610,35 @@ int *givehelp, *giveversion, *verbosity, *printvars;
 		it_came_from_cmdline = 1;
 		if (*(s+1)) {
 		    /* User gave -Mmach */
-		    t = s+1;
+		    mhost = s+1;
 		    s += strlen(s) - 1;
 		} else if ((iarg+1 < argc) && argv[iarg+1][0] != '-') {
 		    /* User gave -M mach */
 		    iarg++;
 		    if (checkarg(argv[iarg]) != 0)
 			return -1;
-		    t = argv[iarg];
+		    mhost = argv[iarg];
 		} else {
 		    return Error(0, 0, "No machine specified after option -M\n");
 		}
-		if (strlen(t) > sizeof(userinfo.hostname)-1)
+		if (strlen(mhost) > sizeof(userinfo.hostname)-1)
 		    Error(1, 1, "Machine name too long -- max allowed is %d\n",
 				sizeof(userinfo.hostname)-1);
-		strcpy(userinfo.hostname, t);
+		strcpy(userinfo.hostname, mhost);
 		if (canonicalize_hostname(userinfo.hostname,
 					    sizeof(userinfo.hostname)) == -1)
-		    return 1;
+		    return -1;
+
+		/* also set lowercase version */
 		strcpy(userinfo.lc_hostname, userinfo.hostname);
 		strtolower(userinfo.lc_hostname);
+
+		/* also adjust short form to match */
+		if ((t = strchr(mhost, '.')))
+		    *t = '\0';
+		if (add_variable("HOST", mhost) == -1)
+		    exit(1);
+
 		break;
 	    case 'T':
 		it_came_from_cmdline = 1;
@@ -602,6 +697,27 @@ int *givehelp, *giveversion, *verbosity, *printvars;
 		    			pw->pw_uid, pw->pw_name, pw->pw_gid);
 		}
 		break;
+	    case 'o':
+		if (*(s+1)) {
+		    /* User gave `-ofile' */
+		    *o_file = s+1;
+		    s += strlen(s) - 1;
+		} else if ((iarg+1 < argc) && argv[iarg+1][0] != '-') {
+		    /* User gave `-o file' */
+		    iarg++;
+		    if (checkarg(argv[iarg]) != 0)
+			return -1;
+		    *o_file = argv[iarg];
+		} else {
+		    return Error(0, 0, "No file specified after option -o\n");
+		}
+		/* The `-o file' is always the end of options -- everything
+		 * else is an argument to the user's command.
+		 * Note that we have to return here (or else "goto") else
+		 * we won't break out of the outer for-loop.
+		 */
+		iarg++;
+		return iarg;
 	    case 'b': *printvars = 1; break;
 	    case 'd':
 		debug = 1;
@@ -623,12 +739,22 @@ int *givehelp, *giveversion, *verbosity, *printvars;
 		    /* User gave -Xfile (where X is -c/-F) */
 		    superfile = s+1;
 		    s += strlen(s) - 1;
+		    if (*o_file) {
+			Error(0, 0, "*** Warning: requested file named \
+by `-o %s' is overridden by file `%s'", *o_file, superfile);
+			*o_file = NULL;
+		    }
 		} else if ((iarg+1 < argc) && argv[iarg+1][0] != '-') {
 		    /* User gave -X file */
 		    iarg++;
 		    if (checkarg(argv[iarg]) != 0)
 			return -1;
 		    superfile = argv[iarg];
+		    if (*o_file) {
+			Error(0, 0, "*** Warning: requested file named \
+by `-o %s' is overridden by file `%s'", *o_file, superfile);
+			*o_file = NULL;
+		    }
 		} else {
 		    /* No file was given.  Use the default super file */
 		}
@@ -695,7 +821,7 @@ char **arglist;
 char **envp;
 int n_builtin;
 {
-    char *s, **p, **sp;
+    char *s, *t, **p, **sp;
     char *cdpath;
     int isglobal, j, iarg;
 
@@ -780,12 +906,12 @@ int n_builtin;
 	fprintf(stderr, "%d chars\n", localinfo.maxenvlen);
     (void) fprintf(stderr,
 	"\tAdditional permitted user's environment variables:\n");
-	if (localinfo.env == NULL || localinfo.env[0] == NULL) {
-	    fprintf(stderr, "\t\t(none)\n");
-	} else {
-	    for (p=localinfo.env; *p; p++)
-		(void) fprintf(stderr, "\t\t%s\n", *p);
-	}
+    if (localinfo.env == NULL || localinfo.env[0] == NULL) {
+	fprintf(stderr, "\t\t(none)\n");
+    } else {
+	for (p=localinfo.env; *p; p++)
+	    (void) fprintf(stderr, "\t\t%s\n", *p);
+    }
 
     (void) fprintf(stderr,
 			"\tEnvironment variables defined with setenv=var=:\n");
@@ -828,74 +954,136 @@ int n_builtin;
     }
 #endif
 
-    (void) fprintf(stderr, "\n\tPassword required: %s\n",
-			    localinfo.passinfo.required ? "yes" : "no");
-    if (localinfo.passinfo.required) {
-	(void) fprintf(stderr, "\tPassword timeout: %d min\n",
-				localinfo.passinfo.timeout);
+    (void) fprintf(stderr, "\n\tAuthentication required: %s\n", auth_name());
+    if (localinfo.authinfo.required) {
+	(void) fprintf(stderr, "\tAuthentication timeout: %d min\n",
+				localinfo.authinfo.timeout);
 	(void) fprintf(stderr,
     "\tUpdate timestamp with each use of a password-requiring command? %s\n",
-				localinfo.passinfo.renewtime ? "Yes" : "No");
+				localinfo.authinfo.renewtime ? "Yes" : "No");
+    }
+
+    (void) fprintf(stderr,
+	"\tVariables that must be entered in response to a prompt:\n");
+    if (localinfo.checkvar == NULL || localinfo.checkvar[0] == NULL) {
+	fprintf(stderr, "\t\t(none)\n");
+    } else {
+	for (p=localinfo.checkvar; *p; p++) {
+	    t = get_variable(*p);
+	    (void) fprintf(stderr, "\t\t%s\t(%s)\n", *p,
+				    t ? t : "** NOT DEFINED! **");
+	}
     }
 
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Switch to processing a user's super file, and set uid/gid/groups
- * to that user.
+ * to that user.  Die if error.
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-int
-user_supertab(command)
-char *command;		/* user:cmd */
+void
+user_supertab(file_or_user, is_file, cmd)
+char *file_or_user;	/* either file from `-o file' or username */
+int is_file;		/* !0 means it's filename, not username */
+char *cmd;		/* for error reporting */
 {
-    char *s, *user, *cmd;
+    char *s, *user;
     struct passwd *pass;
- 
-    /* Split command into user:cmd */
-    if ((cmd = strchr(command, ':')) == NULL)
-	return 0;		/* No ":" -- no user superfile */
+    uid_t uid;
+    struct stat sto, stu;
+    static char no_cmd[] = "";
 
-    user = command;
-    *cmd++ = '\0';
-    if (*user == '\0'  ||  *cmd == '\0')
-	return Error(0, 0, "Commands may not begin or end with `:'\n");
+    if (!cmd)
+	cmd = "";
+ 
+    if (!file_or_user || !*file_or_user)
+	Error(0, 1, "Internal error!  \
+Invoked user_supertab(NULL or \"\", ...)!\n");
+
+    if (is_file) {
+
+	/* Have file; verify it's a .supertab and extract user. */
+	if (stat(file_or_user, &sto) == -1) {
+	    Error(1, 1, "Can't stat file `%s': ", file_or_user);
+	}
+
+	/* Now validate the file: it must refer to the same file
+	 * as the canonical per-user supertab file.
+	 */
+	if (!(pass = getpwuid(sto.st_uid))) {
+	    Error(1, 1, "Can't get passwd entry for uid=%d (owner of `%s'): ",
+	    		sto.st_uid, file_or_user);
+	}
+	/* Construct the user's .supertab filename and put into
+	 * the superfile buffer.
+	 */
+	user = pass->pw_name;
+	(void) construct_user_superfile(user);
+	if (stat(superfile, &stu) == -1) {
+	    Error(1, 1, "Can't stat file `%s': ", superfile);
+	}
+
+	if (stu.st_dev != sto.st_dev || stu.st_ino != sto.st_ino) {
+	    Error(0, 1, "The file `%s' points to a file owned by `%s', \
+but it isn't the expected file `%s'\n",
+		file_or_user, pass->pw_name, superfile);
+	}
+
+    } else {
+	/* Have username */
+	user = file_or_user;
+	pass = construct_user_superfile(user);
+    } 
+    error_srcfile = superfile;
+
+    if (initgroups(user, pass->pw_gid) == -1)
+	Error(1, 1,
+	"Can't set groups to those of user %s, group %d:\n\t\
+Command <%s:%-.500s>; uid=%d, euid=%d: ",
+	    user, pass->pw_gid, user, cmd, getuid(), geteuid());
+
+    if (setgid(pass->pw_gid) == -1)
+	Error(1, 1,
+	    "Can't set gid to that of user %s (from command <%s:%-.500s>): ",
+	    user, user, cmd);
+
+    if (setuid(pass->pw_uid) == -1)
+	Error(1, 1,
+	    "Can't set uid to user %s (from command <%s:%-.500s>): ",
+	    user, user, cmd);
+}
+
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/*
+ * Construct's a user's .supertab file, and place it into the superfile.
+ * Return a pointer to the passwd struct for the user.
+ * Exit if there are any errors.
+ */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+struct passwd *
+construct_user_superfile(user)
+char *user;
+{
+    struct passwd *pass;
+    char *s;
 
     superfile[0] = '\0';
-    pass = getpwnam(user);
-    if (pass == (struct passwd *) NULL)
-	return Error(0, 0, "No such user as `%s' (from command <%s:%s>)\n",
-			user, user, cmd);
- 
+    if (!(pass = getpwnam(user)))
+	Error(0, 1, "No such user as `%s'\n", user);
     if (pass->pw_dir[0] == '\0')
-	return Error(0, 0, "No home directory for user `%s'?!\n", user);
+	Error(0, 1, "No home directory for user `%s'?!\n", user);
     if ((strlen(pass->pw_dir) +
-		strlen(PERUSER_SUPERFILE) + 1) > sizeof(superfile_writable))
-	return Error(0, 0,
-	    "User %s's <home directory>+<%s> pathlen exceeds %d?!\n",
-	    user, PERUSER_SUPERFILE, sizeof(superfile_writable));
+	    strlen(PERUSER_SUPERFILE) + 1) > sizeof(superfile_writable))
+	Error(0, 1,
+		"User %s's <home directory>+<%s> pathlen exceeds %d?!\n",
+		user, PERUSER_SUPERFILE, sizeof(superfile_writable));
     (void) strcpy(superfile, pass->pw_dir);
     s = superfile + strlen(superfile) - 1;
     if (*s != '/')
 	*(++s) = '/';
     strcpy(s+1, PERUSER_SUPERFILE);
-    error_srcfile = superfile;
-
-    if (initgroups(user, pass->pw_gid) == -1)
-	return Error(0, 0,
-	"Can't set groups to those of user %s (from command <%s:%-.500s>).\n",
-	    user, user, cmd);
-
-    if (setgid(pass->pw_gid) == -1)
-	return Error(1, 0,
-	    "Can't set gid to that of user %s (from command <%s:%-.500s>): ",
-	    user, user, cmd);
-
-    if (setuid(pass->pw_uid) == -1)
-	return Error(1, 0,
-	    "Can't set uid to user %s (from command <%s:%-.500s>): ",
-	    user, user, cmd);
-    return 0;
+    return pass;
 }
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -1077,7 +1265,9 @@ char *cmd;		/* name of command being started */
      * USER and HOME refer to the uid under which the command is executed;
      * LOGNAME is set to the same as USER, and SUPERCMD is set to cmd.
      * ORIG_USER, ORIG_LOGNAME, and ORIG_HOME refer to the user who invoked
-     * super.
+     * super (these are values computed by super, not passed in by the caller).
+     * CALLER and CALLER_HOME are set to the same as ORIG_USER and ORIG_HOME,
+     * respectively.
      * Returned:
      *  NULL on error;
      *	otherwise, a pointer to the modified environment list.
@@ -1088,11 +1278,13 @@ char *cmd;		/* name of command being started */
     int fd_log;
     static char *env[200];
     static char User[100];		/* USER */
-    static char Logname[100];		/* LOGNAME (alias for USER) */
+    static char Logname[100];		/* LOGNAME (same as USER) */
     static char Home[MAXPATHLEN+5];	/* HOME */
     static char OrigUser[100];		/* ORIG_USER */
+    static char Caller[100];		/* CALLER (same as ORIG_USER) */
     static char OrigLogname[100];	/* ORIG_LOGNAME */
     static char OrigHome[MAXPATHLEN+9];	/* ORIG_HOME */
+    static char CallerHome[MAXPATHLEN+12]; /* CALLER_HOME (same as ORIG_HOME) */
     static char SafePath[1206];		/* SAFE_PATH */
     static char Cmd[1205];		/* SUPERCMD */
 
@@ -1134,6 +1326,7 @@ char *cmd;		/* name of command being started */
     s = *localinfo.user ? localinfo.user :
 		*localinfo.u_g ? localinfo.u_g : userinfo.caller.pw_name;
     (void) sprintf(OrigUser, "ORIG_USER=%s", userinfo.caller.pw_name);
+    (void) sprintf(Caller, "CALLER=%s", userinfo.caller.pw_name);
     (void) sprintf(User, "USER=%s", s);
     (void) sprintf(OrigLogname, "ORIG_LOGNAME=%s", userinfo.caller.pw_name);
     (void) sprintf(Logname, "LOGNAME=%s", s);
@@ -1159,6 +1352,7 @@ char *cmd;		/* name of command being started */
     (void) strcpy(Home, "HOME=");
     (void) getlogdir(s, Home+5);
     (void) sprintf(OrigHome, "ORIG_HOME=%s", userinfo.caller.pw_dir);
+    (void) sprintf(CallerHome, "CALLER_HOME=%s", userinfo.caller.pw_dir);
     i = 0;
     env[i] = Getenv("TERM");
     if (env[i] && checkenv("TERM", env[i]+5, "^[-/:+._a-zA-Z0-9]*$") != -1) i++;
@@ -1173,8 +1367,10 @@ char *cmd;		/* name of command being started */
     env[i++] = Home;
     env[i++] = Cmd;
     env[i++] = OrigUser;
+    env[i++] = Caller;
     env[i++] = OrigLogname;
     env[i++] = OrigHome;
+    env[i++] = CallerHome;
 
     /* Now add the extra environment variables requested in the
      * super.tab file.  Make a linear search each time, so that we
@@ -1282,20 +1478,33 @@ char **argv;
 void
 add_builtin_variables()
 {
-    char buf[1000];
+    char *s, buf[1000];
 
-    add_variable("HOSTNAME", userinfo.hostname);
+    if (add_variable("HOSTNAME", userinfo.hostname) == -1)
+	exit(1);
 
-    add_sysinfo_variables();	/* Creates empty def'ns if no sysinfo() */
+    /* Add an unqualified version  of HOSTNAME */
+    strcpy(buf, userinfo.hostname);
+    if ((s=strchr(buf, '.'))) {
+	*s = '\0';
+    }
+    if (add_variable("HOST", buf) == -1)
+	exit(1);
 
-    add_uname_variables();	/* Creates empty def'ns if no uname() */
+    if (add_sysinfo_variables() == -1)	/* Makes empty def'ns if no sysinfo() */
+	exit(1);
+
+    if (add_uname_variables() == -1) 	/* Makes empty def'ns if no uname() */
+	exit(1);
 
 #ifdef HAVE_GETDOMAINNAME
     if (getdomainname(buf, sizeof(buf)) != -1) {
-	add_variable("NIS_DOMAIN", buf);
+	if (add_variable("NIS_DOMAIN", buf) == -1)
+	    exit(1);
     }
 #else
-	add_variable("NIS_DOMAIN", "");
+	if (add_variable("NIS_DOMAIN", "") == -1)
+	    exit(1);
 #endif
 
 }
@@ -1336,12 +1545,12 @@ init_localinfo()
     localinfo.fdlist = NULL;
     localinfo.mask = -1;
     localinfo.maxenvlen = globalinfo.maxenvlen;
-    localinfo.env = NULL;
     localinfo.maxlen1arg = globalinfo.maxlen1arg;
     localinfo.maxlenargs = globalinfo.maxlenargs;
     localinfo.file_uid = UID_NOTSET;
     localinfo.file_gid = GID_NOTSET;
     localinfo.nice_incr = globalinfo.nice_incr;
+    localinfo.checkvar = NULL;
 	       
     /* Don't init to global ngroups, because we need to be able to tell
      * later on whether local groups=xxx or global groups=xxx was used.
@@ -1378,7 +1587,7 @@ init_userinfo()
      */
     if (get_canonical_hostname(userinfo.hostname, sizeof(userinfo.hostname))
 	    == -1)
-	return 1;
+	return -1;
     strcpy(userinfo.lc_hostname, userinfo.hostname);
     strtolower(userinfo.lc_hostname);
 
@@ -1390,6 +1599,11 @@ init_userinfo()
 	return Error(0, 0, "Couldn't get your password entry.");
     memcpy(&userinfo.caller, (void *) usrpw, sizeof(struct passwd));
 
+    if (add_variable("CALLER", userinfo.caller.pw_name) == -1)
+	return -1;
+    if (add_variable("CALLER_HOME", userinfo.caller.pw_dir) == -1)
+	return -1;
+
     /* Since the string fields that we need are overwritten by later
      * calls to getpwxxx(), make private copies:
      */
