@@ -161,6 +161,7 @@ void *memset P__(( void *s, int c, size_t n ));
     extern int debug;			/* debug level; 0=off */
     extern int check_syntax;		/* just check syntax of superfile */
     extern char *superfile;		/* The actual superfile to be opened. */
+    extern char superfile_init[];	/* The super.init file. */
     extern int it_came_from_cmdline;	/* Set by -F/-T/-U/-G flags */
     extern int using_user_supertab;	/* !0 means using a user's .supertab */
 
@@ -274,16 +275,26 @@ struct progList {
 typedef struct progList ProgList;
 
 /* ========================================================================= */
+#ifdef HAVE_ENUM
+enum {SUPER_AUTH_PASSWORD, SUPER_AUTH_PAM} Method;
+#else
+/* No enums! */
+#define SUPER_AUTH_PASSWORD 1
+#define SUPER_AUTH_PAM 2
+#endif
 
-/* Password information */
-struct passInfo {
-    int required;	/* -1 means not in use */
-    int timeout;	/* Time before password must be re-entered */
+/* Authentication information */
+struct authInfo {
+    int required;	/* 0 = no auth required */
+    int method;		/* AUTH_PASSWD, AUTH_PAM, etc */
+    int timeout;	/* Time before re-authentication required */
     int renewtime;	/* update the timestamp file with each use of cmd? */
     int perhost;	/* create timestamp files separately for each host? */
-    char user[1024];	/* value of timestampuid=xxx, if entered */
+    char user[1024];	/* value of authuser=xxx, if entered */
+    char ts_user[1024];	/* value of timestampuid=xxx, if entered */
+    char *prompt;	/* optional string with which to prompt for authinfo */
 };
-typedef struct passInfo PassInfo;
+typedef struct authInfo AuthInfo;
 
 /* Information for logging use */
 struct logInfo {
@@ -340,7 +351,7 @@ struct globalInfo {
     long maxlen1arg;	/* max len of any single arg */
     long maxlenargs;	/* max len of all args, combined */
 
-    PassInfo passinfo;	/* password information */
+    AuthInfo authinfo;	/* authentication information */
 
     Simple2List userbefore;	/* list of u/g/h pats before per-cmd pats */
     Simple2List userafter;	/* list of u/g/h pats after per-cmd pats */
@@ -424,12 +435,13 @@ struct localInfo {
     long maxlen1arg;	/* max len of any single arg */
     long maxlenargs;	/* max len of all args, combined */
     StrArray argpats;	/* argNNN=xxx arguments */
+    char **checkvar;	/* null-term'd list of vars from checkvar=var[,...] */
 
     char mailcmd[500];	/* Information for logging via mail */
     int mail_success;	/* bool: mail on successful tries? (-1 = unknown) */
 
     int *fd;		/* descriptors from fdlist string */
-    PassInfo passinfo;	/* password requirements on this command */
+    AuthInfo authinfo;	/* authentication requirements on this command */
 };
 typedef struct localInfo LocalInfo;
 
@@ -448,10 +460,10 @@ extern Conditions matches;		/* To keep track of what matched */
 /* ========================================================================= */
 void	add_builtin_variables P__(( void ));
 #ifdef HAVE_SYSINFO
-void	add_sysinfo_variables P__(( void ));
+int	add_sysinfo_variables P__(( void ));
 #endif
 #ifdef HAVE_UNAME
-void	add_uname_variables P__(( void ));
+int	add_uname_variables P__(( void ));
 #endif
 int	add_variable P__(( char *varname, char *vardefn ));
 void	anchor P__(( char *in, char *out ));
@@ -464,32 +476,37 @@ int	blkprint P__((char **));
 int	build_cmd_file P__((char *, int , char *, char **));
 char**	buttonup P__((char *));
 int	check_arglistlen P__(( char **argv ));
+int	check_auth P__(( char *cmd ));
 int	check_owner P__(( void ));
-int	check_pass P__(( char *cmd ));
 int	check_value P__(( char *value, char *pat ));
+int	check_var_value P__(( void ));
 int	checkenv P__(( char *name,  char *value,  char *pat ));
 char*	clean_buf P__(( char *buf, char *outbuf ));
 void	close_writer P__(( void ));
 int	colon_define P__(( char *command ));
+int	colon_getenv P__(( char *command ));
 int	colon_global P__(( char *command ));
 int	colon_include P__(( char *command, int allow_missing ));
 int	colon_if P__(( char *command, int *colonif_succeeded ));
-int	conditions_and_options P__((char *cond_or_opt_wd));
+int	conditions_and_options P__(( char *cond_or_opt_wd ));
+struct passwd *
+	construct_user_superfile P__(( char *user ));
 char*	dayname P__(( int daynum ));
 int	daynum P__(( int unixtime ));
 void	debug_print P__((char *path, char **arglist, char **envp, int n_bltin));
 void	debug_hello P__((void));
 int	do_options P__((int argc, char **argv, int *help, int *vers,
-					int *verbose, int *printvars));
+			int *verbose, int *printvars, char **owned_by_file));
 char*	do_variables P__(( char *str));
-void	print_variable P__(( int indx, char *key, char *data ));
+char*	docrypt P__((char *key, char *salt));
 char*	ends P__(( char *s1, char *s2 ));
 #ifdef HAVE_STDARG_H
 int	Error P__(( int show_perror, int die, char * fmt, ... ));
 #else
 int	Error();
 #endif
-FileList* file_open P__(( FileList *list, char *name, int allow_missing ));
+FileList* file_open P__(( FileList *list, char *name,
+			int allow_missing, uid_t *req_uid, gid_t *req_gid ));
 FileList* file_close P__(( FileList *list ));
 char*	fillbuffer P__(( FILE *fp, int *indentok, int *nl ));
 int	findgid P__(( int allowbrackets, char *grouplabel ));
@@ -501,9 +518,14 @@ void	free_TimeList P__(( TimeList *));
 int	get_canonical_hostname P__((char *buf, int len));
 int	get_encrypted_pw P__(( void ));
 int	get_owner P__((char *path, uid_t *uid, gid_t *gid));
-int	get_password P__(( char *cmd, char *user,  char *salt,  char *encr ));
+int	get_pam P__(( char *cmd, char *caller, char *user ));
+int	get_password P__(( char *cmd, char *caller, char *user,
+			char *salt,  char *encr ));
+char*	get_variable P__(( char *varname ));
 char*	Getenv P__(( char *s ));
 int	Getgroups P__((int, GETGROUPS_T *));
+struct group *
+	getgrentry P__(( char *name ));
 int	getlogdir P__((char *user, char *buf));
 struct passwd *
 	getpwentry P__((int allow_brackets, char *username));
@@ -533,6 +555,7 @@ int	option_clear_settings P__((char *word, char *s, int isglobal));
 int	option_global_reset_settings P__((void));
 int	option_local_clear_settings P__((void));
 int	parseline P__((int givehelp, int checksyntax, char *buf, char *usrcmd));
+void	print_variable P__(( int indx, char *key, char *data ));
 void	printhelp P__(( int verbose ));
 void	printhelp_hello P__(( int verbose, char *usrcmd ));
 int	process_colon_cmds P__(( char *command, int *colonif_succeeded ));
@@ -545,6 +568,7 @@ int	s_re_exec P__(( char *));
 void	readtime_init P__(( void ));
 int	readtimerange P__(( char *str, short *t1, short *t2, char *d ));
 int	readtime P__(( char *str, short *t1, char *d ));
+int	save_var_value P__((char *var_value));
 int	set_chdir P__((void));
 int	set_nice_incr P__((void));
 int	set_u_g P__(( void ));
@@ -566,4 +590,4 @@ int	StrNelts P__(( StrArray *a ));
 void	strtolower P__(( char *string ));
 char*	strqtokS P__(( char *, char *, char *, char *, unsigned int ));
 int	read_syslogcode P__(( char *str, int *code ));
-int	user_supertab P__(( char *command ));
+void	user_supertab P__(( char *file_or_user, int is_file, char *cmd ));
