@@ -1,8 +1,11 @@
-static char rcsid[] = "$Id: time.c,v 1.26 1998/04/05 18:56:39 will Exp $";
+static char rcsid[] = "$Id: time.c,v 1.27 1999/02/09 05:48:41 will Exp $";
 #include "super.h"
 
 #define ANYDAY 7
 
+/* Maximum length of a time field in super.tab. */
+#define MAX_T_FIELDLEN 80
+
 struct weekDay {
     char name[20];
     char abbr[20];
@@ -294,6 +297,12 @@ char *d;	/* Returned with day number (0=Sunday) or ANYDAY */
     int relop;
 #endif
 
+    if (strlen(str) > MAX_T_FIELDLEN) {
+	return Error(0, 0,
+		"%t\n\tInvalid time range; first %d chars are: <%*.*s>\n",
+		MAX_T_FIELDLEN, MAX_T_FIELDLEN, str);
+    }
+
 #ifdef HAVE_LOCALE_H
     if (*(weekday[0].name) == '\0')
 	readtime_init();
@@ -427,7 +436,7 @@ char *d;	/* Returned with day number (0=Sunday) or ANYDAY */
 	/* It must be a weekday, expressed in the current locale.
 	 * Allow just dayname or /dayname.
 	 */
-	char buf[20];
+	char buf[MAX_T_FIELDLEN+1];
 	strcpy(buf, s);
 	strtolower(buf);
 	for (i=0; i<7; i++) {
@@ -478,7 +487,7 @@ char *d;	/* Returned with day number (0=Sunday) */
 
      */
     char *s, *p;
-    char buf[40];
+    char buf[MAX_T_FIELDLEN+1];
     int hh, mm;
     int i, l;
 
