@@ -1,4 +1,4 @@
-static char rcsid[] = "$Id: utils.c,v 1.97 1999/02/24 05:16:21 will Exp $";
+static char rcsid[] = "$Id: utils.c,v 1.118 2000/08/20 04:55:13 will Exp $";
 #include "super.h"
 
 /* Variables table flag -- !0 means has been created */
@@ -462,23 +462,21 @@ int *nl;		/* Returned with number of lines read in */
  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 FileList *
-file_open(parent, name, allow_missing)
-FileList *parent;
-char *name;
+file_open(parent, name, allow_missing, req_uid, req_gid)
+FileList *parent;	/* null if this is head file */
+char *name;		/* file to open */
 int allow_missing;	/* silently ignore missing files if parent != null. */
+uid_t *req_uid;		/* NULL if default, else ptr to required uid */
+gid_t *req_gid;		/* NULL if default, else ptr to required gid;
+			 * this group is also allowed write permission.
+			 */
 {
     char *s;
 
     FileList *fl;
     struct stat st;
-    static int nobody_uid = -999;
-    static struct passwd *nobodypw;
-    int uid, euid;
-
-    if (nobody_uid == -999) {
-	nobodypw = getpwnam("nobody");
-	nobody_uid = nobodypw ? nobodypw->pw_uid : 65534;
-    }
+    struct passwd *pass;
+    int orig_uid, euid;
 
     fl = (FileList *) malloc(sizeof(FileList));
     if (!fl) {
@@ -519,20 +517,17 @@ int allow_missing;	/* silently ignore missing files if parent != null. */
 	fl->fullname = fl->givenname;
     }
 
-    if (debug)
-	if (parent == NULL)
+    if (debug) {
+	if (parent == NULL) {
 	    fprintf(stderr, "\tOpening file %s\n", fl->fullname);
-	else
+	} else {
 	    fprintf(stderr, "\tLine %d: opening include file %s\n",
 		parent->line, fl->fullname);
+	}
+    }
 
-    /* For any file:
-     *	If we are running as root:
-     *		make sure that the super.tab file is owned by
-     *		root or nobody, and is not writable by group/world.
-     *	Otherwise, if our real uid differs from that under which we
-     *		started the program, make sure the super file is owned
-     *		by ourselves and not group/world writable.
+    /*
+     * Handle missing-file case.
      */
     if (stat(fl->fullname, &st) == -1) {
 	if (errno == ENOENT && allow_missing) {
@@ -552,61 +547,67 @@ int allow_missing;	/* silently ignore missing files if parent != null. */
 	    return NULL;
 	}
     }
+
+    /*
+     * Check file write permissions.
+     */
+    if (st.st_mode & S_IWOTH) {
+	/*
+	 * World-writable is never ok.
+	 */
+	Error(0, 0, "super.tab %sfile `%s' is world-writable.  Bailing out.\n",
+			parent == NULL ? "" : "include ", fl->fullname);
+	if (debug || it_came_from_cmdline) {
+	    Error(0, 0,
+	      "If this weren't debug mode, super would have quit here.\n");
+	} else {
+	    return NULL;
+	}
+
+    } else if (!req_gid && (st.st_mode & S_IWGRP)) {
+	/*
+	 * No group was specified, but it's group-writable!
+	 */
+	Error(0, 0, "super.tab %sfile `%s' is group-writable.  Bailing out.\n",
+			parent == NULL ? "" : "include ", fl->fullname);
+	if (debug || it_came_from_cmdline) {
+	    Error(0, 0,
+	      "If this weren't debug mode, super would have quit here.\n");
+	} else {
+	    return NULL;
+	}
+    }
+
+    if (!parent) {
+	/* This is the master file.  Extract its owner's name for use
+	 * in the SUPER_OWNER variable.
+	 */
+	pass = getpwuid(st.st_uid);
+	if (!pass) {
+	    Error(0, 0, "%t\n\tSuper.tab file `%s' owner uid is %d, \
+which has not passwd entry!\n", fl->fullname, st.st_uid);
+	    return NULL;
+	}
+	if (add_variable("SUPER_OWNER", pass->pw_name) == -1)
+	    return NULL;
+	if (add_variable("SUPER_HOME", pass->pw_dir) == -1)
+	    return NULL;
+    }
+
+    /*
+     * Check file ownership.
+     */
     euid = geteuid();
-    uid = userinfo.orig_uid;
-    if (euid == 0 && uid == 0) {
-	/* Being run by root */
-	if (st.st_uid != 0 && st.st_uid != nobody_uid) {
-	    Error(0, 0, "super.tab %sfile `%s' isn't owned by root \
-or nobody,\n\tbut we are being run by root.  Bailing out.\n",
-		parent == NULL ? "" : "include ", fl->fullname);
-	    if (debug || it_came_from_cmdline) {
-		Error(0, 0,
-		  "If this weren't debug mode, super would have quit here.\n");
-	    } else {
-		return NULL;
-	    }
-	}
-	if (st.st_mode & (S_IWGRP || S_IWOTH)) {
-	    Error(0, 0, "super.tab %sfile `%s' is owned by root \
-or nobody,\n\tbut is group- or world-writeable.  Bailing out.\n",
-		parent == NULL ? "" : "include ", fl->fullname);
-	    if (debug || it_came_from_cmdline) {
-		Error(0, 0,
-		  "If this weren't debug mode, super would have quit here.\n");
-	    } else {
-		return NULL;
-	    }
-	}
-
-    } else if (euid == 0) {
-	/* Running as root, original user someone else */
-	if (st.st_uid != 0 && st.st_uid != nobody_uid) {
-	    Error(0, 0, "super.tab %sfile `%s' isn't owned by root \
-or nobody,\n\tbut our euid==0.   %s to uid=%d.\n",
+    orig_uid = userinfo.orig_uid;
+    if (req_uid && req_gid) {
+	/*
+	 * File's uid/gid must match the required uid/gid.
+	 */
+	if (st.st_uid != *req_uid || st.st_gid != *req_gid) {
+	    Error(0, 0, "super.tab %sfile `%s' needs to have (uid=%d, \
+gid=%d), but is uid=%d, gid=%d\n",
 		parent == NULL ? "" : "include ", fl->fullname,
-		(debug || it_came_from_cmdline) ?
-		    "If this weren't debug mode, we'd revert" : "Reverting",
-		userinfo.orig_uid);
-	    if (!debug)
-		setuid(uid);
-	}
-	if (st.st_mode & (S_IWGRP || S_IWOTH)) {
-	    Error(0, 0, "super.tab %sfile `%s' is owned by root \
-or nobody,\n\tbut is group- or world-writeable.  %s to uid=%d.\n",
-		parent == NULL ? "" : "include ", fl->fullname,
-		(debug || it_came_from_cmdline) ?
-		    "If this weren't debug mode, we'd revert" : "Reverting",
-		userinfo.orig_uid);
-	    if (!debug)
-		setuid(uid);
-	}
-
-    } else if (uid != userinfo.caller.pw_uid) {
-	/* Real uid has changed -- make sure uid is owner of superfile. */
-	if (st.st_uid != uid) {
-	    Error(0, 0, "super.tab %sfile `%s' isn't owned by uid %d.\n",
-		parent == NULL ? "" : "include ", fl->fullname, uid);
+		    *req_uid, *req_gid, st.st_uid, st.st_gid);
 	    if (debug || it_came_from_cmdline) {
 		Error(0, 0,
 		    "If this weren't debug mode, super would quit here.\n");
@@ -614,22 +615,76 @@ or nobody,\n\tbut is group- or world-writeable.  %s to uid=%d.\n",
 		return NULL;
 	    }
 	}
-	if (st.st_mode & (S_IWGRP || S_IWOTH)) {
-	    if (userinfo.orig_uid == 0) {
+
+    } else if (req_uid) {
+	/*
+	 * File's uid must match the required uid.
+	 */
+	if (st.st_uid != *req_uid) {
+	    Error(0, 0,
+		"super.tab %sfile `%s' needs to have uid=%d, but has uid=%d\n",
+		parent == NULL ? "" : "include ", fl->fullname,
+		    *req_uid, st.st_uid);
+	    if (debug || it_came_from_cmdline) {
 		Error(0, 0,
-	    "super.tab %sfile `%s' is group or world-writable.  Bailing out.\n",
-		    parent == NULL ? "" : "include ", fl->fullname);
-		if (debug || it_came_from_cmdline) {
-		    Error(0, 0,
-			"If this weren't debug mode, super would have quit here.\n");
-		} else {
-		    return NULL;
-		}
+		    "If this weren't debug mode, super would quit here.\n");
+	    } else {
+		return NULL;
+	    }
+	}
+
+    } else if (req_gid) {
+	/*
+	 * File's gid must match the required gid.
+	 */
+	if (st.st_gid != *req_gid) {
+	    Error(0, 0,
+		"super.tab %sfile `%s' needs to have gid=%d, but has gid=%d\n",
+		    parent == NULL ? "" : "include ", fl->fullname,
+			*req_gid, st.st_gid);
+	    if (debug || it_came_from_cmdline) {
+		Error(0, 0,
+		    "If this weren't debug mode, super would quit here.\n");
+	    } else {
+		return NULL;
+	    }
+	}
+
+    } else if (euid == 0 || orig_uid == 0) {
+	/* Being run by root, using default rules: the super.tab file
+	 * must be owned by root.
+	 */
+	if (st.st_uid != 0) {
+	    Error(0, 0, "super.tab %sfile `%s' isn't owned by root, \n\
+\tbut we are being run as or by root.  Bailing out.\n",
+		parent == NULL ? "" : "include ", fl->fullname);
+	    if (debug || it_came_from_cmdline) {
+		Error(0, 0,
+		  "If this weren't debug mode, super would have quit here.\n");
+	    } else {
+		return NULL;
+	    }
+	}
+
+    } else if (orig_uid != userinfo.caller.pw_uid) {
+	/*
+	 * Real uid has changed -- this happens when the command is
+	 * like "user:somecommand", in which case the file's owner
+	 * had better be "user".
+	 *
+	 */
+	if (st.st_uid != orig_uid) {
+	    Error(0, 0, "super.tab %sfile `%s' isn't owned by uid %d.\n",
+		parent == NULL ? "" : "include ", fl->fullname, orig_uid);
+	    if (debug || it_came_from_cmdline) {
+		Error(0, 0,
+		    "If this weren't debug mode, super would quit here.\n");
+	    } else {
+		return NULL;
 	    }
 	}
     }
 
-
     if ((fl->fp = fopen(fl->fullname, "r")) == NULL) {
 	Error(1, 0, "%t\n\tCouldn't open super.tab %sfile `%s': ",
 		parent == NULL ? "" : "include ", fl->fullname);
@@ -990,8 +1045,8 @@ char *prefix, *hostname, *file;
     l = strlen(prefix) + 1 + strlen(hostname) + 1;
     if (l >= MAXPATHLEN) {
 	Error(1, 0,
-	"Can't create timestamp directory: it would exceed MAXPATHLEN = %d\n",
-	MAXPATHLEN);
+	"Can't create directory <%.500s...>: it would exceed MAXPATHLEN = %d\n",
+	prefix, MAXPATHLEN);
 	return NULL;
     }
 
@@ -1118,7 +1173,7 @@ char *vardefn;
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Add sysinfo(2) items as variables.  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-void
+int
 add_sysinfo_variables()
 {
     char buf[500];
@@ -1126,79 +1181,96 @@ add_sysinfo_variables()
 #ifdef HAVE_SYSINFO
 #ifdef SI_ARCHITECTURE
     if (sysinfo(SI_ARCHITECTURE, buf, sizeof(buf)) != -1) {
-	add_variable("SI_ARCHITECTURE", buf);
+	if (add_variable("SI_ARCHITECTURE", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_HOSTNAME", "");
+    if (add_variable("SI_HOSTNAME", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_HOSTNAME
     if (sysinfo(SI_HOSTNAME, buf, sizeof(buf)) != -1) {
-	add_variable("SI_HOSTNAME", buf);
+	if (add_variable("SI_HOSTNAME", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_HW_PROVIDER", "");
+    if (add_variable("SI_HW_PROVIDER", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_HW_PROVIDER
     if (sysinfo(SI_HW_PROVIDER, buf, sizeof(buf)) != -1) {
-	add_variable("SI_HW_PROVIDER", buf);
+	if (add_variable("SI_HW_PROVIDER", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_HW_SERIAL", "");
+    if (add_variable("SI_HW_SERIAL", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_HW_SERIAL
     if (sysinfo(SI_HW_SERIAL, buf, sizeof(buf)) != -1) {
-	add_variable("SI_HW_SERIAL", buf);
+	if (add_variable("SI_HW_SERIAL", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_MACHINE", "");
+    if (add_variable("SI_MACHINE", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_MACHINE
     if (sysinfo(SI_MACHINE, buf, sizeof(buf)) != -1) {
-	add_variable("SI_MACHINE", buf);
+	if (add_variable("SI_MACHINE", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_RELEASE", "");
+    if (add_variable("SI_RELEASE", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_RELEASE
     if (sysinfo(SI_RELEASE, buf, sizeof(buf)) != -1) {
-	add_variable("SI_RELEASE", buf);
+	if (add_variable("SI_RELEASE", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_VERSION", "");
+    if (add_variable("SI_VERSION", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_VERSION
     if (sysinfo(SI_VERSION, buf, sizeof(buf)) != -1) {
-	add_variable("SI_VERSION", buf);
+	if (add_variable("SI_VERSION", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_SRPC_DOMAIN", "");
+    if (add_variable("SI_SRPC_DOMAIN", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_SRPC_DOMAIN
     if (sysinfo(SI_SRPC_DOMAIN, buf, sizeof(buf)) != -1) {
-	add_variable("SI_SRPC_DOMAIN", buf);
+	if (add_variable("SI_SRPC_DOMAIN", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
 
-    add_variable("SI_SYSNAME", "");
+    if (add_variable("SI_SYSNAME", "") == -1)
+	return -1;
 #ifdef HAVE_SYSINFO
 #ifdef SI_SYSNAME
     if (sysinfo(SI_SYSNAME, buf, sizeof(buf)) != -1) {
-	add_variable("SI_SYSNAME", buf);
+	if (add_variable("SI_SYSNAME", buf) == -1)
+	    return -1;
     }
 #endif
 #endif
@@ -1208,29 +1280,35 @@ add_sysinfo_variables()
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* Add uname(2) items as variables.  */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-void
+int
 add_uname_variables()
 {
     struct utsname uts;
 
     if (uname(&uts) != -1) {
-	add_variable("UNAME_SYSNAME", uts.sysname);
-	add_variable("UNAME_NODENAME", uts.nodename);
-	add_variable("UNAME_RELEASE", uts.release);
-	add_variable("UNAME_VERSION", uts.version);
-	add_variable("UNAME_MACHINE", uts.machine);
+	if (add_variable("UNAME_SYSNAME", uts.sysname) ||
+	    add_variable("UNAME_NODENAME", uts.nodename) ||
+	    add_variable("UNAME_RELEASE", uts.release) ||
+	    add_variable("UNAME_VERSION", uts.version) ||
+	    add_variable("UNAME_MACHINE", uts.machine))
+		return -1;
     } else {
-	add_variable("UNAME_SYSNAME", "");
-	add_variable("UNAME_NODENAME", "");
-	add_variable("UNAME_RELEASE", "");
-	add_variable("UNAME_VERSION", "");
-	add_variable("UNAME_MACHINE", "");
+	if (add_variable("UNAME_SYSNAME", "") ||
+	    add_variable("UNAME_NODENAME", "") ||
+	    add_variable("UNAME_RELEASE", "") ||
+	    add_variable("UNAME_VERSION", "") ||
+	    add_variable("UNAME_MACHINE", ""))
+		return -1;
     }
 }
 #endif
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-/* Does variable substitution on a string.  */
+/* Does variable substitution on a string.  The result is stored in
+ * an internal, reused buffer, and a pointer to that buffer is returned.
+ * The caller must copy the returned string to a safe place before calling
+ * any routine that might reuse the buffer.
+ */
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 char *
 do_variables(string)
@@ -1363,6 +1441,24 @@ char *string;
     return variablebuf.buf;
 }
 
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+/* Returns a variable's value. */
+/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+char *
+get_variable(varname)
+char *varname;
+{
+    ENTRY wanted_item, *found_item;
+
+    /* Have a variable requiring expansion */
+    wanted_item.key = varname;
+    found_item = hsearch(wanted_item, FIND);
+    if (!found_item) {
+	return NULL;
+    }
+    return found_item->data;
+}
+
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /* A function for printing one variable's name and value.
  * See hprint() for calling sequence.
@@ -1470,7 +1566,7 @@ int len;
 {
 #ifdef HAVE_GETHOSTBYNAME
     struct hostent *he, *he_nodot, *he_dot;
-    char dot_name[80] = "", nodot_name[80] = "";
+    char dot_name[1024] = "", nodot_name[1024] = "";
 
 
     if (globalinfo.gethostbyname) {
